import numpy as np
from math import exp, log10, ceil
import torch
from tqdm import tqdm
import torch.nn.functional as F
from torch.autograd import Variable
from fastai.vision import ImageImageList, Tuple, subplots, plt, random
from .common import ArcGISMSImage, ArcGISImageListRGB, ArcGISImageList
from .._utils.common import get_nbatches, get_top_padding


def gaussian(window_size, sigma):
    gauss = torch.Tensor(
        [
            exp(-((x - window_size // 2) ** 2) / float(2 * sigma**2))
            for x in range(window_size)
        ]
    )
    return gauss / gauss.sum()


def create_window(window_size, channel):
    _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
    _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
    window = Variable(
        _2D_window.expand(channel, 1, window_size, window_size).contiguous()
    )
    return window


def _ssim(img1, img2, window, window_size, channel, size_average=True):
    mu1 = F.conv2d(img1, window, padding=window_size // 2, groups=channel)
    mu2 = F.conv2d(img2, window, padding=window_size // 2, groups=channel)

    mu1_sq = mu1.pow(2)
    mu2_sq = mu2.pow(2)
    mu1_mu2 = mu1 * mu2

    sigma1_sq = (
        F.conv2d(img1 * img1, window, padding=window_size // 2, groups=channel) - mu1_sq
    )
    sigma2_sq = (
        F.conv2d(img2 * img2, window, padding=window_size // 2, groups=channel) - mu2_sq
    )
    sigma12 = (
        F.conv2d(img1 * img2, window, padding=window_size // 2, groups=channel)
        - mu1_mu2
    )

    C1 = 0.01**2
    C2 = 0.03**2

    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / (
        (mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2)
    )

    if size_average:
        return ssim_map.mean()
    else:
        return ssim_map.mean(1).mean(1).mean(1)


class SSIM(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True):
        super(SSIM, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = 1
        self.window = create_window(window_size, self.channel)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.data.type() == img1.data.type():
            window = self.window
        else:
            window = create_window(self.window_size, channel)

            if img1.is_cuda:
                window = window.cuda(img1.get_device())
            window = window.type_as(img1)

            self.window = window
            self.channel = channel

        return _ssim(img1, img2, window, self.window_size, channel, self.size_average)


def ssim(img1, img2, window_size=11, size_average=True):
    (_, channel, _, _) = img1.size()
    window = create_window(window_size, channel)

    if img1.is_cuda:
        window = window.cuda(img1.get_device())
    window = window.type_as(img1)

    return _ssim(img1, img2, window, window_size, channel, size_average)


def psnr(img1, img2):
    mse = F.l1_loss(img1, img2)
    psnr = 10 * log10(1 / mse)
    return psnr


class ArcGISImageListSR_MS(ArcGISImageList):
    _div = None


def standardize(x, means, stds):
    return (x - np.array(means)[:, None, None]) / np.array(stds)[:, None, None]


class ImageImageListSR(ImageImageList):
    label_cls = ArcGISImageListSR_MS

    @classmethod
    def from_folders(cls, path, folder, image_stats, is_multispec=False, **kwargs):
        global _image_stats, is_ms
        res = super().from_folder(folder, **kwargs)
        res.path = path
        _image_stats, is_ms = image_stats, is_multispec
        return res

    def open(self, fn):
        if is_ms:
            return ArcGISMSImage(
                standardize(
                    ArcGISMSImage.open(fn, div=None).data,
                    _image_stats[0],
                    _image_stats[1],
                ).type(torch.float32)
            )
        else:
            return ArcGISMSImage.open(fn, div=255)

    def label_from_func(self, func, label_cls=None, **kwargs):
        if is_ms:
            label_cls = ArcGISImageListSR_MS
        else:
            label_cls = ArcGISImageListRGB
        return super().label_from_func(func, label_cls=label_cls, **kwargs)


def show_batch(self, rows=4, **kwargs):
    """
    This function randomly picks a few training chips and visualizes them.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional int. Number of rows of results
                            to be displayed.
    =====================   ===========================================
    """
    xs, ys = [], []
    for n, imgs in enumerate(self.train_ds):
        if n != rows:
            xs.append(imgs[0])
            ys.append(imgs[1])
        else:
            break

    axs = subplots(len(xs), 2, figsize=(20, rows * 5))
    for i, (x, y) in enumerate(zip(xs, ys)):
        x, y = ArcGISMSImage(x.data), ArcGISMSImage(y.data)
        x.show(ax=axs[i, 0], **kwargs)
        y.show(ax=axs[i, 1], **kwargs)
    axs[0, 0].title.set_text("Low Resolution")
    axs[0, 1].title.set_text("High Resolution")


def show_results(self, rows, **kwargs):
    from .._data_utils.pix2pix_data import display_row, denormalize

    sampling = kwargs.get("sampling_type", "ddim")
    ntimestep = kwargs.get("n_timestep")
    device = next(self.learn.model.parameters()).device.type

    self.learn.model.eval()
    activ = []
    top = get_top_padding(title_font_size=16, nrows=rows, imsize=5)
    denormfunc = lambda img, max, min: (img + 1) * (max - min) / 2 + min

    if self.model_type == "UNet":
        x_batch, y_batch = get_nbatches(
            self._data.valid_dl, ceil(rows / self._data.batch_size)
        )
        if isinstance(x_batch[0], list):
            x_batch = x_batch[0]

        x_A, x_B = torch.cat(x_batch), torch.cat(y_batch)

        for i in range(0, x_A.shape[0], self._data.batch_size):
            preds = self.learn.model(x_A[i : i + self._data.batch_size].detach())
            activ.append(preds)
        activations = torch.cat(activ)

        x_A = denormalize(x_A.cpu(), *self._data._image_stats)
        x_B = denormalize(x_B.cpu(), *self._data._image_stats)
        activations = denormalize(activations.cpu(), *self._data._image_stats)
        rows = min(rows, x_A.shape[0])
    else:
        x_A = torch.cat([i[0] for i, _ in self.learn.data.valid_dl])[:rows]
        x_B = torch.cat([j for _, j in self.learn.data.valid_dl])[:rows]

        x_A_batch = x_A.detach()

        if sampling == "ddim":
            n_timestep = ntimestep if ntimestep else 200
            nstp = {"n_timestep": self.kwargs.get("n_timestep", 1000)}
        else:
            n_timestep = ntimestep if ntimestep else self.kwargs.get("n_timestep", 1000)
            nstp = {"n_timestep": n_timestep}
        combkwargs = {**kwargs, **self.kwargs, **nstp}

        self.learn.model.set_new_noise_schedule(device, **combkwargs)

        preds = []
        for k in tqdm(
            range(x_A_batch.shape[0]), desc="sampling loop time step per image"
        ):
            if sampling == "ddim":
                preds.append(
                    self.learn.model.super_resolution(
                        x_A_batch[k, None],
                        continous=False,
                        sampling_timesteps=n_timestep,
                        ddim_sampling_eta=1,
                        sampling="ddim",
                    )
                )
            else:
                preds.append(
                    self.learn.model.super_resolution(
                        x_A_batch[k, None], continous=False
                    )
                )
        activations = torch.cat(preds)

        maxvals_a = self._data.batch_stats_a["band_max_values"][..., None, None]
        minvals_a = self._data.batch_stats_a["band_min_values"][..., None, None]
        maxvals_b = self._data.batch_stats_b["band_max_values"][..., None, None]
        minvals_b = self._data.batch_stats_b["band_min_values"][..., None, None]
        x_A = denormfunc(x_A.cpu(), maxvals_a, minvals_a)
        x_B = denormfunc(x_B.cpu(), maxvals_b, minvals_b)
        activations = denormfunc(activations.cpu(), maxvals_b, minvals_b)

    fig, axs = plt.subplots(
        nrows=rows, ncols=3, figsize=(4 * 5, rows * 5), squeeze=False
    )
    plt.subplots_adjust(top=top)
    axs[0, 0].title.set_text("Input")
    axs[0, 1].title.set_text("Target")
    axs[0, 2].title.set_text("Prediction")
    for r in range(rows):
        display_row(
            axs[r],
            (
                ArcGISMSImage(x_A[r].cpu()),
                ArcGISMSImage(x_B[r].cpu()),
                ArcGISMSImage(activations[r].detach().cpu()),
            ),
            kwargs.get("rgb_bands", None),
        )



from .._layer import ImageryLayer, Raster, RasterCollection, _ArcpyRasterCollection
from arcgis.gis import Item
import numbers
from arcgis.features.layer import FeatureLayer
from arcgis.auth import (
    EsriUserTokenAuth,
    EsriOAuth2Auth,
    EsriNotebookAuth,
    EsriAPIKeyAuth,
    ArcGISProAuth,
    EsriBuiltInAuth,
    EsriGenTokenAuth,
)
import requests

_aggregating_functions = [
    "max",
    "min",
    "med",
    "mean",
    "majority",
    "sum",
    "std",
    "variety",
    "geometric_median",
    "minority",
    "merge_rasters",
]


def _raster_input(raster, raster2=None):
    layer = None

    # if input is a rastercollection, get the list of rasters and use that as the input

    if isinstance(raster, RasterCollection) or isinstance(raster2, RasterCollection):
        import inspect

        fn_name_l2 = inspect.stack()[2][3]
        fn_name_l1 = inspect.stack()[1][3]
        if (
            fn_name_l2 != "to_multidimensional_raster"
            and fn_name_l1 != "_simple_collection"
        ):
            if (
                not fn_name_l2 in _aggregating_functions
                and not fn_name_l1 in _aggregating_functions
            ):
                raise RuntimeError(
                    "RasterCollection object cannot be specified as input to non aggregating functions"
                )
            if isinstance(raster, RasterCollection):
                if (
                    hasattr(raster, "_ras_coll_engine")
                ) and raster._ras_coll_engine != _ArcpyRasterCollection:
                    raster = raster._ras_coll_engine_obj._rasters_list
                else:
                    if hasattr(raster, "_ras_coll_engine_obj"):
                        return (
                            raster._ras_coll_engine_obj,
                            raster._ras_coll_engine_obj,
                            raster._ras_coll_engine_obj,
                        )
                    else:
                        return raster, raster, raster

            if (
                isinstance(raster2, RasterCollection)
            ) and raster2._ras_coll_engine != _ArcpyRasterCollection:
                if (
                    hasattr(raster2, "_ras_coll_engine")
                ) and raster2._ras_coll_engine != _ArcpyRasterCollection:
                    raster2 = raster2._ras_coll_engine_obj._rasters_list
                else:
                    if hasattr(raster2, "_ras_coll_engine_obj"):
                        return (
                            raster2._ras_coll_engine_obj,
                            raster2._ras_coll_engine_obj,
                            raster2._ras_coll_engine_obj,
                        )
                    else:
                        return raster2, raster2, raster2

    if isinstance(raster, Raster):
        if hasattr(raster, "_engine_obj"):
            raster = raster._engine_obj
    if isinstance(raster2, Raster):
        if hasattr(raster2, "_engine_obj"):
            raster2 = raster2._engine_obj
    if isinstance(raster, list):
        for index, ele in enumerate(raster):
            if isinstance(ele, Raster):
                if hasattr(ele, "_engine_obj"):
                    raster[index] = ele._engine_obj
    if isinstance(raster2, list):
        for index, ele in enumerate(raster2):
            if isinstance(ele, Raster):
                if hasattr(ele, "_engine_obj"):
                    raster2[index] = ele._engine_obj

    if raster2 is not None:
        if isinstance(raster2, (ImageryLayer, Raster)) and isinstance(
            raster, (ImageryLayer, Raster)
        ):
            if raster2._rendering_rule_from_item:
                raster2._fn = None
                raster2._fnra = None
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            if raster._datastore_raster and raster2._datastore_raster:
                if raster2._fn is not None:
                    if raster._uri == raster2._uri:
                        raster2 = raster2._fn
                    else:
                        raster2 = _replace_raster_url(raster2._fn, raster2._uri)
                else:
                    if raster2._uri == raster._uri:
                        oids = raster2.filtered_rasters()
                        if oids is None:
                            raster2 = "$$"
                        elif len(oids) == 1:
                            raster2 = "$" + str(oids[0])
                        else:
                            raster2 = ["$" + str(x) for x in oids]
                    else:
                        raster2 = raster2._uri
            else:
                if raster2._fn is not None:
                    if raster._url == raster2._url:
                        raster2 = raster2._fn
                    else:
                        if raster2._datastore_raster is False:
                            try:
                                url = raster2._url
                                if (
                                    (hasattr(raster2, "_lazy_token"))
                                    and raster2._lazy_token is None
                                ) or not hasattr(raster2, "_lazy_token"):
                                    raster2._lazy_token = _generate_layer_token(
                                        raster2, url
                                    )
                                if isinstance(raster2._lazy_token, str):
                                    url = url + "?token=" + raster2._lazy_token
                                raster2 = _replace_raster_url(raster2._fn, url)
                            except:
                                if "url" in raster2._lyr_dict:
                                    url = raster2._lyr_dict["url"]
                                    if "serviceToken" in raster2._lyr_dict:
                                        url = (
                                            url
                                            + "?token="
                                            + raster2._lyr_dict["serviceToken"]
                                        )
                                    raster2 = _replace_raster_url(raster2._fn, url)
                                else:
                                    raster2 = _replace_raster_url(
                                        raster2._fn, raster2._url
                                    )
                        else:
                            raster2 = _replace_raster_url(raster2._fn, raster2._uri)
                else:
                    if raster2._url == raster._url:
                        oids = raster2.filtered_rasters()
                        if oids is None:
                            raster2 = "$$"
                        elif len(oids) == 1:
                            raster2 = "$" + str(oids[0])
                        else:
                            raster2 = ["$" + str(x) for x in oids]
                    else:
                        if raster2._datastore_raster is False:
                            try:
                                url = raster2._url
                                if (
                                    (hasattr(raster2, "_lazy_token"))
                                    and raster2._lazy_token is None
                                ) or not hasattr(raster2, "_lazy_token"):
                                    raster2._lazy_token = _generate_layer_token(
                                        raster2, url
                                    )

                                if isinstance(raster2._lazy_token, str):
                                    url = url + "?token=" + raster2._lazy_token
                                raster2 = url
                            except:
                                if "url" in raster2._lyr_dict:
                                    url = raster2._lyr_dict["url"]
                                    if "serviceToken" in raster2._lyr_dict:
                                        url = (
                                            url
                                            + "?token="
                                            + raster2._lyr_dict["serviceToken"]
                                        )
                                    raster2 = url
                                else:
                                    raster2 = raster2._url
                        else:
                            raster2 = raster2._uri
        elif isinstance(raster2, (ImageryLayer, Raster)) and not isinstance(
            raster, (ImageryLayer, Raster)
        ):
            layer = raster2
            raster_ra = _get_raster_ra(raster2)
            raster2 = _get_raster(raster2)

        elif isinstance(raster2, list):
            mix_and_match = False  # mixing rasters from two image services
            # try:
            #     r0 = raster2[0]
            #     r1 = raster2[1]
            #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
            #         mix_and_match = True
            # except:
            #     pass

            for r in raster2:  # layer is first non numeric raster in list
                if not isinstance(r, numbers.Number):
                    layer = r
                    break

            for r in raster2:
                if r._datastore_raster:
                    if not isinstance(r, numbers.Number):
                        if r._uri != layer._uri:
                            mix_and_match = True
                else:
                    if not isinstance(r, numbers.Number):
                        if r._url != layer._url:
                            mix_and_match = True

            raster_ra = [_get_raster_ra(r) for r in raster2]
            if mix_and_match:
                raster2 = [_get_raster_url(r, layer) for r in raster2]
            else:
                raster2 = [_get_raster(r) for r in raster2]
        else:  # secondinput maybe scalar for arithmetic functions, or a chained raster fn
            layer = None
            # raster = raster
            raster_ra = raster2
        return layer, raster2, raster_ra

    if isinstance(raster, (ImageryLayer, Raster)):
        if raster._rendering_rule_from_item:
            raster._fn = None
            raster._fnra = None
        layer = raster
        raster_ra = _get_raster_ra(raster)
        raster = _get_raster(raster)

    elif isinstance(raster, list):
        mix_and_match = False  # mixing rasters from two image services
        # try:
        #     r0 = raster[0]
        #     r1 = raster[1]
        #     if r0._fn is None and r1._fn is None and r0._url != r1._url:
        #         mix_and_match = True
        # except:
        #     pass

        for r in raster:
            if isinstance(r, (ImageryLayer, Raster)):
                if r._datastore_raster:
                    layer = r
        if layer is None:
            for r in raster:  # layer is first non numeric raster in list
                if not isinstance(r, numbers.Number):
                    layer = r
                    break

        for r in raster:
            if not isinstance(r, numbers.Number):
                if r._datastore_raster:
                    if r._uri != layer._uri:
                        mix_and_match = True
                else:
                    if r._url != layer._url:
                        mix_and_match = True

        raster_ra = [_get_raster_ra(r) for r in raster]
        if mix_and_match:
            raster = [_get_raster_url(r, layer) for r in raster]
        else:
            raster = [_get_raster(r) for r in raster]
    else:  # maybe scalar for arithmetic functions, or a chained raster fn
        layer = None
        # raster = raster
        raster_ra = raster

    return layer, raster, raster_ra


def _get_raster(raster):
    if isinstance(raster, (ImageryLayer, Raster)):
        if raster._fn is not None:
            raster = raster._fn
        else:
            oids = raster.filtered_rasters()
            if oids is None and raster.mosaic_rule is not None:
                raster = _get_raster_ra(raster)
            elif oids is None:
                raster = "$$"
            elif len(oids) == 1:
                raster = "$" + str(oids[0])
            else:
                raster = ["$" + str(x) for x in oids]
    return raster


def _replace_raster_url(obj, url=None):
    # replace all "Raster" : '$$' with url
    if isinstance(obj, dict):
        value = {k: _replace_raster_url(v, url) for k, v in obj.items()}
    elif isinstance(obj, list):
        value = [_replace_raster_url(elem, url) for elem in obj]
    else:
        value = obj

    if value == "$$":
        return url
    elif isinstance(value, str) and len(value) > 0 and value[0] == "$":
        return url + "/" + value.replace("$", "")
    else:
        return value


def _get_raster_url(raster, layer):
    if isinstance(raster, (ImageryLayer, Raster)):
        if raster._fn is not None:
            if raster._datastore_raster and layer._datastore_raster:
                if raster._uri == layer._uri or isinstance(raster._uri, bytes):
                    raster = raster._fn
                else:
                    raster = _replace_raster_url(raster._fn, raster._uri)

            else:
                if raster._url == layer._url:
                    raster = raster._fn
                else:
                    try:
                        url = raster._url
                        if (
                            (hasattr(raster, "_lazy_token"))
                            and raster._lazy_token is None
                        ) or not hasattr(raster, "_lazy_token"):
                            raster._lazy_token = _generate_layer_token(raster, url)
                        if isinstance(raster._lazy_token, str):
                            url = url + "?token=" + raster._lazy_token
                        raster = _replace_raster_url(raster._fn, url)
                    except:
                        if "url" in raster._lyr_dict:
                            url = raster._lyr_dict["url"]
                            if "serviceToken" in raster._lyr_dict:
                                url = url + "?token=" + raster._lyr_dict["serviceToken"]
                            raster = _replace_raster_url(raster._fn, url)
                        else:
                            raster = _replace_raster_url(raster._fn, raster._url)

        else:
            if raster._datastore_raster and layer._datastore_raster:
                if raster._uri == layer._uri:
                    raster = "$$"
                else:
                    raster = raster._uri
            else:
                if raster._url == layer._url:
                    raster = "$$"
                else:
                    try:
                        url = raster._url
                        if (
                            (hasattr(raster, "_lazy_token"))
                            and raster._lazy_token is None
                        ) or not hasattr(raster, "_lazy_token"):
                            raster._lazy_token = _generate_layer_token(raster, url)
                        if isinstance(raster._lazy_token, str):
                            url = url + "?token=" + raster._lazy_token
                        raster = url
                    except:
                        if "url" in raster._lyr_dict:
                            url = raster._lyr_dict["url"]
                            if "serviceToken" in raster._lyr_dict:
                                url = url + "?token=" + raster._lyr_dict["serviceToken"]
                            raster = url
                        else:
                            raster = raster._url

            # oids = raster.filtered_rasters()
            # if oids is None:
            #     raster = '$$'
            # elif len(oids) == 1:
            #     raster = '$' + str(oids[0])
            # else:
            #     raster = ['$' + str(x) for x in oids]
    return raster


def _get_raster_ra(raster):
    if isinstance(raster, (ImageryLayer, Raster)):
        try:
            url = raster._url
            if (
                (hasattr(raster, "_lazy_token")) and raster._lazy_token is None
            ) or not hasattr(raster, "_lazy_token"):
                raster._lazy_token = _generate_layer_token(raster, url)
            if isinstance(raster._lazy_token, str):
                url = url + "?token=" + raster._lazy_token
        except:
            if "url" in raster._lyr_dict:
                url = raster._lyr_dict["url"]
            if "serviceToken" in raster._lyr_dict:
                url = url + "?token=" + raster._lyr_dict["serviceToken"]
        if raster._fnra is not None:
            raster_ra = raster._fnra
        else:
            raster_ra = {}
            if raster._mosaic_rule is not None:
                if raster._datastore_raster:
                    raster_ra["uri"] = raster._uri
                else:
                    raster_ra["url"] = url
                raster_ra["mosaicRule"] = raster._mosaic_rule
            else:
                if raster._datastore_raster:
                    raster_ra = raster._uri

                else:
                    raster_ra = url

            # if raster._mosaic_rule is not None:
            #    raster_ra['mosaicRule'] = raster._mosaic_rule
    elif isinstance(raster, Item):
        raise RuntimeError(
            "Item not supported as input. Use ImageryLayer - e.g. item.layers[0]"
        )
        # raster_ra = {
        #    'itemId': raster.itemid
        # }
    else:
        raster_ra = raster

    return raster_ra


def _raster_input_rft(raster, raster2=None):
    if isinstance(raster, (ImageryLayer, Raster)) or isinstance(raster, FeatureLayer):
        raster_ra = _get_raster_ra_rft(raster)

    elif isinstance(raster, list):
        raster_ra = [_get_raster_ra_rft(r) for r in raster]

    else:  # maybe scalar for arithmetic functions, or a chained raster fn
        raster_ra = raster

    return raster_ra


def _get_raster_ra_rft(raster):
    if isinstance(raster, Raster):
        if hasattr(raster, "_engine_obj"):
            raster = raster._engine_obj
    if isinstance(raster, (ImageryLayer, Raster)):
        if raster._rendering_rule_from_item:
            raster._fn = None
            raster._fnra = None
        try:
            url = raster._url
            if (
                (hasattr(raster, "_lazy_token")) and raster._lazy_token is None
            ) or not hasattr(raster, "_lazy_token"):
                raster._lazy_token = _generate_layer_token(raster, url)
            if isinstance(raster._lazy_token, str):
                url = url + "?token=" + raster._lazy_token
        except:
            if "url" in raster._lyr_dict:
                url = raster._lyr_dict["url"]
            if "serviceToken" in raster._lyr_dict:
                url = url + "?token=" + raster._lyr_dict["serviceToken"]
        if raster._fnra is not None:
            raster_ra = raster._fnra
        else:
            raster_ra = {}
            if raster._mosaic_rule is not None:
                if raster._datastore_raster:
                    raster_ra["uri"] = raster._uri
                else:
                    raster_ra["url"] = url
                raster_ra["mosaicRule"] = raster._mosaic_rule
            else:
                if raster._datastore_raster:
                    raster_ra = raster._uri
                else:
                    raster_ra = url

    elif isinstance(raster, FeatureLayer):
        raster_ra = raster._url
        # if raster._mosaic_rule is not None:
        #    raster_ra['mosaicRule'] = raster._mosaic_rule
    elif isinstance(raster, Item):
        raise RuntimeError(
            "Item not supported as input. Use ImageryLayer - e.g. item.layers[0]"
        )
    else:
        raster_ra = raster

    return raster_ra


def _input_rft(input_layer):
    if isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        if (
            "/fileShares/" in input_layer
            or "/rasterStores/" in input_layer
            or "/cloudStores/" in input_layer
            or ("http" not in input_layer and "https" not in input_layer)
        ):
            input_param = {"uri": input_layer}
        else:
            input_param = {"url": input_layer}

    elif isinstance(input_layer, list):
        input_param = []
        for il in input_layer:
            if isinstance(il, str):
                if (
                    "/fileShares/" in il
                    or "/rasterStores/" in il
                    or "/cloudStores/" in il
                ):
                    input_param.append({"uri": il})
                else:
                    input_param.append({"url": il})
            elif isinstance(il, dict):
                input_param.append(il)
    else:
        input_param = input_layer
    return input_param


def _find_object_ref(rft_dict, record, instance):
    for k, v in rft_dict.items():
        if isinstance(v, dict):
            if "_object_id" in v:
                record[v["_object_id"]] = v
            if "isPublic" in v:
                if v["isPublic"] is True:
                    instance._is_public_flag = True
            _find_object_ref(v, record, instance)

        elif isinstance(v, list):
            for ele in v:
                if isinstance(ele, dict):
                    if "_object_id" in ele:
                        record[ele["_object_id"]] = ele
                    if "isPublic" in ele:
                        if ele["isPublic"] is True:
                            instance._is_public_flag = True
                    _find_object_ref(ele, record, instance)
    return _replace_object_id(rft_dict, record)


def _replace_object_id(rft_dict, record):
    if isinstance(rft_dict, dict):
        if "_object_ref_id" in rft_dict.keys():
            ref_value = record[rft_dict["_object_ref_id"]]
            rft_dict = ref_value
            return _replace_object_id(rft_dict, record)
        else:
            for k, v in rft_dict.items():
                if isinstance(v, dict):
                    rft_dict[k] = _replace_object_id(v, record)

                elif isinstance(v, list):
                    for n, ele in enumerate(v):
                        if isinstance(ele, dict):
                            v[n] = _replace_object_id(ele, record)
    return rft_dict


def _python_variable_name(var):
    var = "".join(e for e in var if e.isalnum() or e == "_")
    return var


def _pixel_type_string_to_long(pixel_type):
    pixel_type = pixel_type.upper()
    if pixel_type == "U1":
        return 0
    elif pixel_type == "U2":
        return 1
    elif pixel_type == "U4":
        return 2
    elif pixel_type == "U8":
        return 3
    elif pixel_type == "S8":
        return 4
    elif pixel_type == "U16":
        return 5
    elif pixel_type == "S16":
        return 6
    elif pixel_type == "U32":
        return 7
    elif pixel_type == "S32":
        return 8
    elif pixel_type == "F32":
        return 9
    elif pixel_type == "F64":
        return 10
    elif pixel_type == "C16":
        return 13
    elif pixel_type == "C32":
        return 14
    elif pixel_type == "C64":
        return 11
    elif pixel_type == "C128":
        return 12
    elif pixel_type == "D8":
        return 15
    elif pixel_type == "D10":
        return 16
    elif pixel_type == "D12":
        return 17
    elif pixel_type == "D16":
        return 18
    elif pixel_type == "D20":
        return 19
    elif pixel_type == "D24":
        return 20
    elif pixel_type == "D32":
        return 21
    elif pixel_type == "D36":
        return 22
    else:
        return -1


def _generate_layer_token(layer, url):
    token = layer._gis._con._create_token(url)
    if token:
        if isinstance(
            layer._gis._con._session.auth,
            (
                EsriUserTokenAuth,
                EsriOAuth2Auth,
                EsriNotebookAuth,
                EsriAPIKeyAuth,
                ArcGISProAuth,
                EsriBuiltInAuth,
                EsriGenTokenAuth,
            ),
        ):
            temp_r = requests.get(
                url, {"f": "json", "token": token}, verify=layer._gis._verify_cert
            )
            if temp_r.status_code != 200:
                return None
            elif (
                temp_r.status_code == 200
                and temp_r.text.lower().find("invalid token") > -1
            ):
                return None
            else:
                return token
        else:
            return token
    else:
        return None


def _set_multidimensional_rules(function_chain=None, function_chain_ra=None):
    from ... import env

    match_variables = env.match_variables
    union_dimension = env.union_dimension

    if (match_variables is not None) and isinstance(match_variables, bool):
        if (
            function_chain is not None
        ) and "MatchVariable" not in function_chain.keys():
            function_chain["rasterFunctionArguments"]["MatchVariable"] = match_variables
        if (
            function_chain_ra is not None
        ) and "MatchVariable" not in function_chain_ra.keys():
            function_chain_ra["rasterFunctionArguments"][
                "MatchVariable"
            ] = match_variables
    if (union_dimension is not None) and isinstance(union_dimension, bool):
        if (
            function_chain is not None
        ) and "UnionDimension" not in function_chain.keys():
            function_chain["rasterFunctionArguments"][
                "UnionDimension"
            ] = union_dimension
        if (
            function_chain_ra is not None
        ) and "UnionDimension" not in function_chain.keys():
            function_chain_ra["rasterFunctionArguments"][
                "UnionDimension"
            ] = union_dimension


def _get_dimension_names(lyr):
    dim_list = []
    if isinstance(lyr, Raster):
        if hasattr(lyr, "_engine_obj"):
            lyr = lyr._engine_obj
    md_info = lyr.multidimensional_info
    if md_info:
        for ele in md_info["multidimensionalInfo"]["variables"]:
            for ele_dim in ele["dimensions"]:
                dim_list.append(ele_dim["name"])
    return dim_list



from ._base import BaseEsriAuth
from ._pki import EsriPKIAuth
from ._winauth import EsriKerberosAuth, EsriWindowsAuth
from ._negotiate import EsriHttpNegotiateAuth
from ._ntlm import EsriHttpNtlmAuth
from ._apikey import EsriAPIKeyAuth
from ._provided_token import (
    EsriUserTokenAuth,
)  # should be used for NBS server and user provided tokens
from ._basic import EsriBasicAuth
from ._token import (
    EsriBuiltInAuth,
    EsriGenTokenAuth,
    ArcGISProAuth,
    ArcGISServerAuth,
)  # legacy or for stand alone only
from ._oauth import EsriOAuth2Auth
from ._notebook import EsriNotebookAuth
from ._pkce import EsriPKCEAuth
from ._utils import check_response_for_error

__all__ = [
    "EsriAPIKeyAuth",
    "EsriBasicAuth",
    "EsriBuiltInAuth",
    "EsriGenTokenAuth",
    "EsriKerberosAuth",
    "EsriNotebookAuth",
    "EsriOAuth2Auth",
    "EsriPKIAuth",
    "EsriUserTokenAuth",
    "ArcGISProAuth",
    "EsriWindowsAuth",
    "BaseEsriAuth",
    "EsriPKCEAuth",
    "ArcGISServerAuth",
    "EsriHttpNtlmAuth",
    "EsriHttpNegotiateAuth",
    "check_response_for_error",
]



import zipfile
import os

__all__ = ["_zipdir"]


#  --------------------------------------------------------------------
def _zipdir(src, dst, zip_name):
    """
    Function creates zip archive from src in dst location. The name of archive is zip_name.
    :param src: Path to directory to be archived.
    :param dst: Path where archived dir will be stored.
    :param zip_name: The name of the archive.
    :return: None
    """

    zip_name = os.path.join(dst, zip_name)
    ### zipfile handler
    ziph = zipfile.ZipFile(zip_name, "w")
    ### writing content of src directory to the archive
    for root, dirs, files in os.walk(src):
        for file in files:
            ziph.write(
                os.path.join(root, file),
                arcname=os.path.join(root.replace(src, ""), file),
            )
    ziph.close()
    return zip_name



from .temporal_encoder_decoder import (
    ConvTransformerTokensToEmbeddingNeck,
    TemporalViTEncoder,
    GeospatialNeck,
    TemporalEncoderDecoder,
)

__all__ = [
    "TemporalViTEncoder",
    "ConvTransformerTokensToEmbeddingNeck",
    "TemporalEncoderDecoder",
    "GeospatialNeck",
]



from ._arcgis_model import ArcGISModel
from ._arcgis_model import _EmptyData, _change_tail

HAS_OPENCV = True
try:
    from pathlib import Path
    import warnings
    import json
    import math
    import PIL
    import torch
    import torch.nn as nn
    from collections import OrderedDict
    from fastai.vision.learner import create_body
    from fastai.callbacks.hooks import num_features_model
    from fastai.vision import flatten_model
    from fastai.vision.image import pil2tensor
    from fastai.core import split_kwargs_by_func
    from torchvision import models
    from skimage.measure import find_contours
    from ._codetemplate import instance_detector_prf
    from arcgis.learn.models._deepsort_predict_utils import non_max_suppression
    import numpy as np
    import types
    from ._arcgis_model import (
        _get_backbone_meta,
        _resnet_family,
        _set_ddp_multigpu,
        _isnotebook,
    )
    from ._timm_utils import timm_config, filter_timm_models, _get_feature_size
    from torchvision import models
    from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
    from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
    from fastai.basic_train import Learner
    from ._maskrcnn_utils import (
        is_no_color,
        mask_rcnn_loss,
        train_callback,
        compute_class_AP,
        predict_tta,
        AveragePrecision,
        forward_roi,
        postprocess_transform,
        post_nms_top_n,
        pre_nms_top_n,
        eager_outputs_modified,
    )
    from .._image_utils import _get_image_chips, _draw_predictions

    from ._MaskRCNN_PointRend import create_pointrend
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from fastai.torch_core import split_model_idx
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    import matplotlib
    from torchvision.models.detection.backbone_utils import resnet_fpn_backbone
    from .._utils.common import get_nbatches, image_batch_stretcher, read_image
    from .._utils.env import is_arcgispronotebook

    HAS_FASTAI = True
except Exception as e:
    # raise Exception(e)
    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False


class TimmFPNBackbone(nn.Module):
    def __init__(self, backbone, chip_size=224):
        from fastai.callbacks.hooks import hook_outputs
        from fastai.callbacks.hooks import model_sizes
        from torchvision.ops.feature_pyramid_network import (
            FeaturePyramidNetwork,
            LastLevelMaxPool,
        )
        from ._hed_utils import get_hooks

        super().__init__()
        self.backbone = backbone
        hooks = get_hooks(self.backbone, chip_size)
        self.hook = hook_outputs(hooks[1:])
        model_sizes(self.backbone, size=(chip_size, chip_size))
        layer_num_channels = [k.stored.shape[1] for k in self.hook]
        self.fpn = FeaturePyramidNetwork(
            in_channels_list=layer_num_channels,
            out_channels=256,
            extra_blocks=LastLevelMaxPool(),
        )
        self.out_channels = 256

    def forward(self, x):
        x = self.backbone(x)
        out = OrderedDict()
        features = self.hook.stored
        for k, v in enumerate(features):
            out[str(k)] = v
        x = self.fpn(out)
        return x


def chips_to_batch(chips, model_height, model_width, batch_size=1):
    dtype = np.float32
    band_count = 3
    if len(chips) != 0:
        dtype = chips[0].dtype

    batch = np.zeros(
        shape=(batch_size, band_count, model_height, model_width),
        dtype=dtype,
    )
    for b in range(batch_size):
        if b < len(chips):
            batch[b, :, :model_height, :model_height] = chips[b]

    return batch


def grid_anchors(self, grid_sizes, strides):
    anchors = []
    cell_anchors = self.cell_anchors
    assert cell_anchors is not None

    for size, stride, base_anchors in zip(grid_sizes, strides, cell_anchors):
        grid_height, grid_width = size
        stride_height, stride_width = stride
        device = base_anchors.device

        # For output anchor, compute [x_center, y_center, x_center, y_center]
        shifts_x = (
            torch.arange(0, grid_width, dtype=torch.float32, device=device)
            * stride_width
        )
        shifts_y = (
            torch.arange(0, grid_height, dtype=torch.float32, device=device)
            * stride_height
        )
        shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x)
        shift_x = shift_x.reshape(-1)
        shift_y = shift_y.reshape(-1)
        shifts = torch.stack((shift_x, shift_y, shift_x, shift_y), dim=1)

        # For every (base anchor, output anchor) pair,
        # offset each zero-centered base anchor by the center of the output anchor.
        anchors.append(
            (shifts.view(-1, 1, 4) + base_anchors.view(1, -1, 4)).reshape(-1, 4)
        )

    return anchors


class MaskRCNNTracer(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.model.eval()

    def _dict_to_tuple(self, out_dict):
        return (
            out_dict["boxes"],
            out_dict["scores"],
            out_dict["labels"],
            out_dict["masks"],
        )

    def forward(self, inp):
        out = self.model(inp)
        return self._dict_to_tuple(out[0])


class MaskRCNN(ArcGISModel):
    """
    Model architecture from https://arxiv.org/abs/1703.06870.
    Creates a :class:`~arcgis.learn.MaskRCNN` Instance segmentation model,
    based on https://github.com/pytorch/vision/blob/master/torchvision/models/detection/mask_rcnn.py.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone convolutional neural network
                            model used for feature extraction, which
                            is `resnet50` by default.
                            Supported backbones: ResNet family and specified Timm
                            models(experimental support) from :func:`~arcgis.learn.MaskRCNN.backbones`.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    pointrend               Optional boolean. If True, it will use PointRend
                            architecture on top of the segmentation head.
                            Default: False. PointRend architecture from
                            https://arxiv.org/pdf/1912.08193.pdf.
    =====================   ===========================================

    **kwargs**

    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_train         Optional int. Number of proposals to keep before
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_test          Optional int. Number of proposals to keep before
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_train        Optional int. Number of proposals to keep after
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_test         Optional int. Number of proposals to keep after
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_nms_thresh                  Optional float. NMS threshold used for postprocessing
                                    the RPN proposals.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_fg_iou_thresh               Optional float. Minimum IoU between the anchor
                                    and the GT box so that they can be considered
                                    as positive during training of the RPN.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_bg_iou_thresh               Optional float. Maximum IoU between the anchor and
                                    the GT box so that they can be considered as negative
                                    during training of the RPN.
                                    Default: 0.3
    -----------------------------   ---------------------------------------------
    rpn_batch_size_per_image        Optional int. Number of anchors that are sampled
                                    during training of the RPN for computing the loss.
                                    Default: 256
    -----------------------------   ---------------------------------------------
    rpn_positive_fraction           Optional float. Proportion of positive anchors in a
                                    mini-batch during training of the RPN.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_score_thresh                Optional float. During inference, only return proposals
                                    with a classification score greater than box_score_thresh
                                    Default: 0.05
    -----------------------------   ---------------------------------------------
    box_nms_thresh                  Optional float. NMS threshold for the prediction head.
                                    Used during inference.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_detections_per_img          Optional int. Maximum number of detections per
                                    image, for all classes.
                                    Default: 100
    -----------------------------   ---------------------------------------------
    box_fg_iou_thresh               Optional float. Minimum IoU between the proposals and
                                    the GT box so that they can be considered as positive
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_bg_iou_thresh               Optional float. Maximum IoU between the proposals and
                                    the GT box so that they can be considered as negative
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_batch_size_per_image        Optional int. Number of proposals that are sampled during
                                    training of the classification head.
                                    Default: 512
    -----------------------------   ---------------------------------------------
    box_positive_fraction           Optional float. Proportion of positive proposals in a
                                    mini-batch during training of the classification head.
                                    Default: 0.25
    =============================   =============================================

    :return:
        :class:`~arcgis.learn.MaskRCNN` Object
    """

    def __init__(
        self,
        data,
        backbone=None,
        pretrained_path=None,
        pointrend=False,
        *args,
        **kwargs,
    ):
        # Set default backbone to be 'resnet50'
        if backbone is None:
            backbone = models.resnet50

        if pretrained_path is not None:
            pretrained_backbone = False
        else:
            pretrained_backbone = True

        self._check_dataset_support(data)
        if not (self._check_backbone_support(backbone)):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)
        if self._is_multispectral:
            self._backbone_ms = self._backbone
            self._backbone = self._orig_backbone
            scaled_mean_values = data._scaled_mean_values[data._extract_bands].tolist()
            scaled_std_values = data._scaled_std_values[data._extract_bands].tolist()

        self._code = instance_detector_prf

        self._pointrend = pointrend

        self.maskrcnn_kwargs, kwargs = split_kwargs_by_func(
            kwargs, models.detection.MaskRCNN.__init__
        )

        if (
            self._backbone.__name__ == "resnet50"
            and "timm" not in self._backbone.__module__
        ):
            model = models.detection.maskrcnn_resnet50_fpn(
                pretrained=pretrained_backbone,
                pretrained_backbone=False,
                min_size=1.5 * data.chip_size,
                max_size=2 * data.chip_size,
                **self.maskrcnn_kwargs,
            )

            if self._is_multispectral:
                model.backbone = _change_tail(model.backbone, data)
                model.transform.image_mean = scaled_mean_values
                model.transform.image_std = scaled_std_values
        elif (
            self._backbone.__name__ in ["resnet18", "resnet34"]
            and not pointrend
            and "timm" not in self._backbone.__module__
        ):
            if self._is_multispectral:
                backbone_small = create_body(
                    self._backbone_ms,
                    pretrained=pretrained_backbone,
                    cut=_get_backbone_meta(self._backbone.__name__)["cut"],
                )
                backbone_small.out_channels = 512
                model = models.detection.MaskRCNN(
                    backbone_small,
                    91,
                    min_size=1.5 * data.chip_size,
                    max_size=2 * data.chip_size,
                    image_mean=scaled_mean_values,
                    image_std=scaled_std_values,
                    **self.maskrcnn_kwargs,
                )
            else:
                backbone_small = create_body(
                    self._backbone, pretrained=pretrained_backbone
                )
                backbone_small.out_channels = 512
                model = models.detection.MaskRCNN(
                    backbone_small,
                    91,
                    min_size=1.5 * data.chip_size,
                    max_size=2 * data.chip_size,
                    **self.maskrcnn_kwargs,
                )
            model.rpn.anchor_generator.grid_anchors = types.MethodType(
                grid_anchors, model.rpn.anchor_generator
            )
        else:
            if "timm" in self._backbone.__module__:
                backbone_cut = timm_config(self._backbone)["cut"]
                backbone_fpn = create_body(
                    self._backbone, pretrained_backbone, backbone_cut
                )
                try:
                    backbone_fpn = TimmFPNBackbone(backbone_fpn, data.chip_size)
                except:
                    if "tresnet" in self._backbone.__module__:
                        backbone_fpn.out_channels = _get_feature_size(
                            self._backbone, backbone_cut
                        )[-1][1]
                    else:
                        backbone_fpn.out_channels = num_features_model(
                            torch.nn.Sequential(*backbone_fpn.children())
                        )
            else:
                backbone_fpn = resnet_fpn_backbone(
                    self._backbone.__name__, pretrained=pretrained_backbone
                )
            if self._is_multispectral:
                backbone_fpn = _change_tail(backbone_fpn, data)
                model = models.detection.MaskRCNN(
                    backbone_fpn,
                    91,
                    min_size=1.5 * data.chip_size,
                    max_size=2 * data.chip_size,
                    image_mean=scaled_mean_values,
                    image_std=scaled_std_values,
                    **self.maskrcnn_kwargs,
                )
            else:
                model = models.detection.MaskRCNN(
                    backbone_fpn,
                    91,
                    min_size=1.5 * data.chip_size,
                    max_size=2 * data.chip_size,
                    **self.maskrcnn_kwargs,
                )
            if "timm" in self._backbone.__module__:
                model.rpn.anchor_generator.grid_anchors = types.MethodType(
                    grid_anchors, model.rpn.anchor_generator
                )

        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, data.c)

        if pointrend:
            model = create_pointrend(model, data.c)
        else:
            in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
            hidden_layer = 256
            model.roi_heads.mask_predictor = MaskRCNNPredictor(
                in_features_mask, hidden_layer, data.c
            )

        if not _isnotebook():
            _set_ddp_multigpu(self)
            if self._multigpu_training:
                self.learn = Learner(
                    data, model, loss_func=mask_rcnn_loss
                ).to_distributed(self._rank_distributed)
                self._map_location = {"cuda:%d" % 0: "cuda:%d" % self._rank_distributed}
            else:
                self.learn = Learner(data, model, loss_func=mask_rcnn_loss)
        else:
            self.learn = Learner(data, model, loss_func=mask_rcnn_loss)
        self.learn.callbacks.append(train_callback(self.learn))
        if not pointrend:
            self.learn.model.roi_heads.forward = types.MethodType(
                forward_roi, self.learn.model.roi_heads
            )
        self.learn.model.eager_outputs = types.MethodType(
            eager_outputs_modified, self.learn.model
        )
        self.learn.model.transform.postprocess = types.MethodType(
            postprocess_transform, self.learn.model.transform
        )
        self.learn.model.rpn.post_nms_top_n = types.MethodType(
            post_nms_top_n, self.learn.model.rpn
        )
        self.learn.model.rpn.pre_nms_top_n = types.MethodType(
            pre_nms_top_n, self.learn.model.rpn
        )
        self.learn.metrics = [AveragePrecision(self.learn)]
        self.learn.model = self.learn.model.to(self._device)
        self.learn.c_device = self._device

        # fixes for zero division error when slice is passed
        idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

        # make first conv weights learnable
        self._arcgis_init_callback()

        if pretrained_path is not None:
            self.load(pretrained_path)

        if hasattr(data, "_image_space_used"):
            if data._image_space_used == "MAP_SPACE":
                self.learn.model.arcgis_tta = list(range(8))
            else:
                self.learn.model.arcgis_tta = [0, 2]
        if self._is_multispectral:
            self._orig_backbone = self._backbone
            self._backbone = self._backbone_ms

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue
            for p in i.parameters():
                p.requires_grad = False
        return idx

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return MaskRCNN._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return MaskRCNN._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(["*repvgg*", "*tresnet*"])
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))
        return [*_resnet_family] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return MaskRCNN._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["RCNN_Masks"]

    @classmethod
    def from_model(cls, emd_path, data=None, **kwargs):
        """
        Creates a :class:`~arcgis.learn.MaskRCNN` Instance segmentation object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.MaskRCNN` Object
        """

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        maskrcnn_kwargs = emd.get("MaskRCNNkwargs", {})

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=kwargs.get("chip_size", emd["ImageHeight"]),
            )
            data.resize_to = emd.get("resize_to", None)
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)

        return cls(
            data, **model_params, pretrained_path=str(model_file), **maskrcnn_kwargs
        )

    def _save_pytorch_torchscript(self, name, save=True):
        model = self.learn.model
        model.eval()

        cpu = torch.device("cpu")
        device = cpu
        model = model.to(cpu)

        if hasattr(self._data, "chip_size"):
            chip_size = self._data.chip_size
            if not isinstance(chip_size, tuple):
                chip_size = (chip_size, chip_size)
        inp = torch.rand(1, 3, chip_size[0], chip_size[1]).to(cpu)

        model = MaskRCNNTracer(model)

        traced_model = None
        try:
            with torch.no_grad():
                traced_model = self._trace(model, inp, True)
        except:
            with torch.no_grad():
                traced_model = self._trace(model, inp, True)

        model.to(device)
        saved_path_cpu = (
            self.learn.path / self.learn.model_dir / f"{name}-cpu.pt"
        ).__str__()
        saved_path_gpu = ""
        torch.jit.save(traced_model, saved_path_cpu)

        if not save:
            [traced_model, traced_model]
        return [f"{name}-cpu.pt", saved_path_gpu]

    def _save_pytorch_tflite(self, name):
        import tensorflow as tf
        import logging

        tf.get_logger().setLevel(logging.ERROR)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import onnx
            import onnx_tf
            from onnx_tf.backend import prepare

        traced_models = self._save_pytorch_torchscript(name, False)
        if traced_models[0] is None:
            return ["", ""]

        cpu = torch.device("cpu")
        device = cpu
        traced_model = traced_models[0].to(device)

        if hasattr(self._data, "chip_size"):
            chip_size = self._data.chip_size
            if not isinstance(chip_size, tuple):
                chip_size = (chip_size, chip_size)
        num_input_channels = list(self.learn.model.parameters())[0].shape[1]
        inp = torch.randn([1, num_input_channels, chip_size[0], chip_size[1]]).to(
            device
        )

        save_path_tflite = self.learn.path / self.learn.model_dir / f"{name}.tflite"
        save_path_onnx = self.learn.path / self.learn.model_dir / f"{name}.onnx"
        save_path_pb = self.learn.path / self.learn.model_dir / f"{name}"

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            torch.onnx.export(
                traced_model,
                inp,
                save_path_onnx,
                export_params=True,
                do_constant_folding=False,
                verbose=True,
                input_names=["input"],
                output_names=["output"],
                opset_version=11,
            )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            arcgis_onnx = onnx.load(save_path_onnx)
            tf_onnx = prepare(arcgis_onnx, logging_level="ERROR")
            tf_onnx.export_graph(str(save_path_pb))

        model = tf.saved_model.load(save_path_pb)
        concrete_func = model.signatures[
            tf.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY
        ]
        converter = tf.lite.TFLiteConverter.from_concrete_functions([concrete_func])
        converter.optimizations = [tf.lite.Optimize.DEFAULT]  # for full accuracy
        tf_lite_model = converter.convert()
        open(save_path_tflite, "wb").write(tf_lite_model)

        return [save_path_tflite, save_path_onnx]

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {"ModelParameters": {}}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_maskrcnn_inferencing"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISInstanceDetector.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISInstanceDetector.py"
        _emd_template["ModelType"] = "InstanceDetection"
        _emd_template["MaskRCNNkwargs"] = self.maskrcnn_kwargs
        _emd_template["ModelParameters"]["pointrend"] = self._pointrend
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["SupportsVariableTileSize"] = True
        _emd_template["Classes"] = []
        class_data = {}
        for i, class_name in enumerate(
            self._data.classes[1:]
        ):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = (
                [random.choice(range(256)) for i in range(3)]
                if is_no_color(self._data.color_mapping)
                else self._data.color_mapping[inverse_class_mapping[class_name]]
            )
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    @property
    def _model_metrics(self):
        return {
            "average_precision_score": self.average_precision_score(show_progress=True)
        }

    def _predict_batch(self, images, detect_thresh=0.5, tta_prediction=False):
        model = self.learn.model
        model.eval()
        model = model.to(self._device)
        normed_batch_tensor = images.to(self._device)
        if tta_prediction:
            predictions = predict_tta(model, normed_batch_tensor, detect_thresh)
        else:
            temp = model.roi_heads.score_thresh
            model.roi_heads.score_thresh = detect_thresh
            predictions = model(list(normed_batch_tensor))
            model.roi_heads.score_thresh = temp
        normed_batch_tensor.detach().cpu()
        del normed_batch_tensor
        return predictions

    def _predict_results(self, xb, detect_thresh=0.5, tta_prediction=False):
        predictions = self._predict_batch(xb, detect_thresh, tta_prediction)
        predictionsf = []
        for i in range(len(predictions)):
            predictionsf.append({})
            predictionsf[i]["masks"] = predictions[i]["masks"].detach().cpu().numpy()
            predictionsf[i]["boxes"] = predictions[i]["boxes"].detach().cpu().numpy()
            predictionsf[i]["labels"] = predictions[i]["labels"].detach().cpu().numpy()
            predictionsf[i]["scores"] = predictions[i]["scores"].detach().cpu().numpy()
            del predictions[i]["masks"]
            del predictions[i]["boxes"]
            del predictions[i]["labels"]
            del predictions[i]["scores"]
        if self._device == torch.device("cuda"):
            torch.cuda.empty_cache()
        return predictionsf

    def _predict_postprocess(self, predictions, threshold=0.5, box_threshold=0.5):
        pred_mask = []
        pred_box = []

        for i in range(len(predictions)):
            out = predictions[i]["masks"].squeeze()
            pred_box.append([])

            if out.shape[0] != 0:  # handle for prediction with n masks
                if (
                    len(out.shape) == 2
                ):  # for out dimension hxw (in case of only one predicted mask)
                    out = out[None]
                ymask = np.where(out[0] > threshold, 1, 0)
                if predictions[i]["scores"][0] > box_threshold:
                    pred_box[i].append(predictions[i]["boxes"][0])
                for j in range(1, out.shape[0]):
                    ym1 = np.where(out[j] > threshold, j + 1, 0)
                    ymask += ym1
                    if predictions[i]["scores"][j] > box_threshold:
                        pred_box[i].append(predictions[i]["boxes"][j])
            else:
                ymask = np.zeros(
                    (self._data.chip_size, self._data.chip_size)
                )  # handle for not predicted masks
            pred_mask.append(ymask)
        return pred_mask, pred_box

    def show_results(
        self,
        rows=4,
        mode="mask",
        mask_threshold=0.5,
        box_threshold=0.7,
        tta_prediction=False,
        imsize=5,
        index=0,
        alpha=0.5,
        cmap="tab20",
        **kwargs,
    ):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        mode                    Required arguments within ['bbox', 'mask', 'bbox_mask'].
                                    * ``bbox`` - For visualizing only bounding boxes.
                                    * ``mask`` - For visualizing only mask
                                    * ``bbox_mask`` - For visualizing both mask and bounding boxes.
        ---------------------   -------------------------------------------
        mask_threshold          Optional float. The probability above which
                                a pixel will be considered mask.
        ---------------------   -------------------------------------------
        box_threshold           Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        tta_prediction          Optional bool. Perform test time augmentation
                                while predicting
        =====================   ===========================================
        """
        self._check_requisites()
        if mode not in ["bbox", "mask", "bbox_mask"]:
            raise Exception("mode can be only ['bbox', 'mask', 'bbox_mask']")

        # Get Number of items
        nrows = rows
        ncols = 2

        type_data_loader = kwargs.get(
            "data_loader", "validation"
        )  # options : traininig, validation, testing
        if type_data_loader == "training":
            data_loader = self._data.train_dl
        elif type_data_loader == "validation":
            data_loader = self._data.valid_dl
        elif type_data_loader == "testing":
            data_loader = self._data.test_dl
        else:
            e = Exception(
                f"could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing "
            )
            raise (e)

        statistics_type = kwargs.get(
            "statistics_type", "dataset"
        )  # Accepted Values `dataset`, `DRA`
        stretch_type = kwargs.get(
            "stretch_type", "minmax"
        )  # Accepted Values `minmax`, `percentclip`

        cmap_fn = getattr(matplotlib.cm, cmap)
        return_fig = kwargs.get("return_fig", False)

        x_batch, y_batch = get_nbatches(data_loader, nrows)
        x_batch = torch.cat(x_batch)
        y_batch = torch.cat(y_batch)

        nrows = min(nrows, len(x_batch))

        title_font_size = 16
        if kwargs.get("top", None) is not None:
            top = kwargs.get("top")
        else:
            top = 1 - (math.sqrt(title_font_size) / math.sqrt(100 * nrows * imsize))

        # Get Predictions
        prediction_store = []
        for i in range(0, x_batch.shape[0], self._data.batch_size):
            prediction_store.extend(
                self._predict_results(
                    x_batch[i : i + self._data.batch_size],
                    box_threshold,
                    tta_prediction,
                )
            )
        pred_mask, pred_box = self._predict_postprocess(
            prediction_store, mask_threshold, box_threshold
        )

        if self._is_multispectral:
            rgb_bands = kwargs.get("rgb_bands", self._data._symbology_rgb_bands)

            e = Exception(
                "`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1."
            )
            symbology_bands = []
            if not (len(rgb_bands) == 3 or len(rgb_bands) == 1):
                raise (e)
            for b in rgb_bands:
                if type(b) == str:
                    b_index = self._bands.index(b)
                elif type(b) == int:
                    self._bands[
                        b
                    ]  # To check if the band index specified by the user really exists.
                    b_index = b
                else:
                    raise (e)
                b_index = self._data._extract_bands.index(b_index)
                symbology_bands.append(b_index)

            # Denormalize X
            if self._data._do_normalize:
                x_batch = (
                    self._data._scaled_std_values[self._data._extract_bands]
                    .view(1, -1, 1, 1)
                    .to(x_batch)
                    * x_batch
                ) + self._data._scaled_mean_values[self._data._extract_bands].view(
                    1, -1, 1, 1
                ).to(
                    x_batch
                )

            # Extract RGB Bands
            symbology_x_batch = x_batch[:, symbology_bands]
            if stretch_type is not None:
                symbology_x_batch = image_batch_stretcher(
                    symbology_x_batch, stretch_type, statistics_type
                )

            # Channel first to channel last for plotting
            symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
            # Clamp float values to range 0 - 1
            if symbology_x_batch.mean() < 1:
                symbology_x_batch = symbology_x_batch.clamp(0, 1)
        else:
            symbology_x_batch = x_batch.permute(0, 2, 3, 1)

        # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
        if symbology_x_batch.shape[-1] == 1:
            symbology_x_batch = symbology_x_batch.squeeze()

        fig, ax = plt.subplots(
            nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
        )
        fig.suptitle("Ground Truth / Predictions", fontsize=title_font_size)
        for i in range(nrows):
            if nrows == 1:
                ax_i = ax
            else:
                ax_i = ax[i]

            # Ground Truth
            ax_i[0].imshow(symbology_x_batch[i].cpu())
            ax_i[0].axis("off")
            if mode in ["mask", "bbox_mask"]:
                n_instance = y_batch[i].unique().shape[0]
                y_merged = y_batch[i].max(dim=0)[0].cpu().numpy()
                try:
                    y_rgba = cmap_fn.resampled(n_instance)(y_merged)
                except:
                    y_rgba = cmap_fn._resample(n_instance)(y_merged)
                y_rgba[y_merged == 0] = 0
                y_rgba[:, :, -1] = alpha
                ax_i[0].imshow(y_rgba)
            ax_i[0].axis("off")

            # Predictions
            ax_i[1].imshow(symbology_x_batch[i].cpu())
            ax_i[1].axis("off")
            if mode in ["mask", "bbox_mask"]:
                n_instance = np.unique(pred_mask[i]).shape[0]
                try:
                    p_rgba = cmap_fn.resampled(n_instance)(pred_mask[i])
                except:
                    p_rgba = cmap_fn._resample(n_instance)(pred_mask[i])
                p_rgba[pred_mask[i] == 0] = 0
                p_rgba[:, :, -1] = alpha
                ax_i[1].imshow(p_rgba)
            if mode in ["bbox_mask", "bbox"]:
                if pred_box[i] != []:
                    for num_boxes in pred_box[i]:
                        rect = patches.Rectangle(
                            (num_boxes[0], num_boxes[1]),
                            num_boxes[2] - num_boxes[0],
                            num_boxes[3] - num_boxes[1],
                            linewidth=1,
                            edgecolor="r",
                            facecolor="none",
                        )
                        ax_i[1].add_patch(rect)
            ax_i[1].axis("off")
        plt.subplots_adjust(top=top)
        if self._device == torch.device("cuda"):
            torch.cuda.empty_cache()

        if is_arcgispronotebook():
            plt.show()
        if return_fig:
            return fig

    def average_precision_score(
        self,
        detect_thresh=0.5,
        iou_thresh=0.5,
        mean=False,
        show_progress=True,
        tta_prediction=False,
    ):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth mask, above
                                which a predicted mask will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        ---------------------   -------------------------------------------
        tta_prediction          Optional bool. Perform test time augmentation
                                while predicting
        =====================   ===========================================
        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        if mean:
            aps = compute_class_AP(
                self,
                self._data.valid_dl,
                1,
                show_progress,
                detect_thresh,
                iou_thresh,
                mean,
                tta_prediction,
            )
            return aps
        else:
            aps = compute_class_AP(
                self,
                self._data.valid_dl,
                self._data.c - 1,
                show_progress,
                detect_thresh,
                iou_thresh,
                tta_prediction=tta_prediction,
            )
            return dict(zip(self._data.classes[1:], aps))

    def _get_transformed_image(self, image, resize=False):
        if resize:
            resize_to = None
            if self._data.chip_size is not None:
                resize_to = self._data.resize_to
            elif self._data.resize_to is not None:
                resize_to = self._data.chip_size

            if resize_to is not None:
                if isinstance(resize_to, tuple):
                    image = cv2.resize(image, resize_to)
                else:
                    image = cv2.resize(image, (resize_to, resize_to))
        return image

    def _pixel_mask_image(
        self,
        batch,
        orig_img_dim,
        min_obj_size,
        threshold,
        offsets,
        scale_ratio,
        pred_mask,
        pred_box,
        pred_class,
        pred_score,
        extra_chips=0,
        tta_prediction=False,
    ):
        predictions = self._predict_batch(
            torch.tensor(batch).float(), threshold, tta_prediction
        )

        for batch_idx in range(len(predictions) - extra_chips):
            offset = offsets[batch_idx]
            masks = predictions[batch_idx]["masks"].squeeze().detach().cpu().numpy()
            if masks.shape[0] != 0:
                if len(masks.shape) == 2:
                    masks = masks[None]

                for n, mask in enumerate(masks):
                    if predictions[batch_idx]["scores"][n].tolist() >= threshold:
                        contours = find_contours(mask, 0.5, fully_connected="high")
                        coord_list = []
                        for c_idx, contour in enumerate(contours):
                            contour[:, 0] = (contour[:, 0] + (offset[0])) * scale_ratio[
                                0
                            ]
                            contour[:, 1] = (contour[:, 1] + (offset[1])) * scale_ratio[
                                1
                            ]
                            if c_idx == 0:
                                coord_list.append(contour[:, [1, 0]].tolist())
                            else:
                                coord_list.append(
                                    list(reversed(contour[:, [1, 0]].tolist()))
                                )
                        box = (
                            predictions[batch_idx]["boxes"][n]
                            .cpu()
                            .detach()
                            .numpy()
                            .tolist()
                        )

                        box[0] = (box[0] + offset[1]) * scale_ratio[1]
                        box[2] = (box[2] + offset[1]) * scale_ratio[1]
                        box[1] = (box[1] + offset[0]) * scale_ratio[0]
                        box[3] = (box[3] + offset[0]) * scale_ratio[0]

                        if box[0] < 0:
                            box[0] = 0

                        if box[1] < 0:
                            box[1] = 0

                        if box[2] >= orig_img_dim[1]:
                            box[2] = orig_img_dim[1] - 1

                        if box[3] >= orig_img_dim[0]:
                            box[3] = orig_img_dim[0] - 1

                        box[2] -= box[0]
                        box[3] -= box[1]
                        area = box[2] * box[3]
                        if area > 0 and math.sqrt(area) >= min_obj_size:
                            pred_box.append(box)
                            pred_class.append(
                                predictions[batch_idx]["labels"][n].tolist()
                            )
                            pred_score.append(
                                predictions[batch_idx]["scores"][n].tolist()
                            )
                            pred_mask.append(coord_list)

    def _get_batched_predictions(
        self,
        chips,
        tytx,
        orig_img_dim,
        scale_ratio,
        threshold=0.5,
        nms_overlap=0.3,
        min_obj_size=1,
        batch_size=1,
        tta_prediction=False,
    ):
        data = []
        offsets = []

        pred_mask = []
        pred_box = []
        pred_class = []
        pred_score = []

        data_counter = 0
        for idx in range(len(chips)):
            chip = chips[idx]
            if self._data._is_multispectral:
                t = torch.tensor(
                    np.rollaxis(chip["chip"], -1, 0).astype(np.float32),
                    dtype=torch.float32,
                )[None]
                scaled_t = self._data._min_max_scaler(t)[0]
                frame = scaled_t[self._data._extract_bands].detach().cpu().numpy()
            else:
                frame = (
                    pil2tensor(
                        PIL.Image.fromarray(
                            cv2.cvtColor(chip["chip"], cv2.COLOR_BGR2RGB)
                        ),
                        dtype=np.float32,
                    )
                    .div_(255)
                    .detach()
                    .cpu()
                    .numpy()
                )

            data.append(frame)
            offsets.append((chip["ymin"], chip["xmin"]))

            data_counter += 1

            if data_counter % batch_size == 0 or idx == len(chips) - 1:
                batch = chips_to_batch(data, tytx, tytx, batch_size)

                self._pixel_mask_image(
                    batch,
                    orig_img_dim,
                    min_obj_size,
                    threshold,
                    offsets,
                    scale_ratio,
                    pred_mask=pred_mask,
                    pred_box=pred_box,
                    pred_class=pred_class,
                    pred_score=pred_score,
                    extra_chips=batch_size - len(data),
                    tta_prediction=tta_prediction,
                )
                data = []
                offsets = []
                data_counter = 0

        import copy

        pred_box_indices = non_max_suppression(
            copy.deepcopy(np.array(pred_box)), nms_overlap, pred_score
        )
        filtered_mask = [pred_mask[i] for i in pred_box_indices]
        filtered_box = [pred_box[i] for i in pred_box_indices]
        filtered_class = [
            self._data.class_mapping[pred_class[i]] for i in pred_box_indices
        ]
        filtered_score = [pred_score[i] for i in pred_box_indices]

        return filtered_mask, filtered_box, filtered_class, filtered_score

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=True,
        visualize=False,
        resize=False,
        tta_prediction=False,
        **kwargs,
    ):
        """
        Predicts and displays the results of a trained model on a single image.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean.
                                Will return the probability scores of the
                                bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        ---------------------   -------------------------------------------
        tta_prediction          Optional bool. Perform test time augmentation
                                while predicting
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        batch_size              Optional int. Batch size to be used
                                during tiled inferencing
        ---------------------   -------------------------------------------
        min_obj_size            Optional int. Minimum object size
                                to be detected.
        =====================   ===========================================

        :return: 'List' of xmin, ymin, width, height, labels, scores, of predicted bounding boxes on the given image
        """

        if not HAS_OPENCV:
            raise Exception(
                "This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24"
            )

        if isinstance(image_path, str):
            import os

            if not os.path.isfile(image_path):
                return None, None, None
            if self._data._is_multispectral:
                resize_to = None
                if resize:
                    if self._data.resize_to is not None:
                        resize_to = self._data.resize_to
                    elif self._data.chip_size is not None:
                        resize_to = self._data.chip_size
                image = read_image(image_path, resize_to)
                resize = False
            else:
                image = cv2.imread(image_path)
        else:
            image = image_path

        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()

        image = self._get_transformed_image(image, resize)
        height, width, _ = image.shape

        tytx = int(kwargs.get("tile_size", self._data.chip_size))
        batch_size = int(kwargs.get("batch_size", 1))
        min_obj_size = int(kwargs.get("min_obj_size", 1))

        if not resize:
            chips = _get_image_chips(image, tytx)
        else:
            chips = [
                {
                    "width": width,
                    "height": height,
                    "xmin": 0,
                    "ymin": 0,
                    "chip": image,
                    "predictions": [],
                }
            ]

        masks, predictions, labels, scores = self._get_batched_predictions(
            chips,
            tytx,
            (orig_height, orig_width),
            (orig_height / (1.0 * height), orig_width / (1.0 * width)),
            threshold,
            nms_overlap,
            min_obj_size,
            batch_size,
            tta_prediction,
        )

        if visualize:
            if self._data._is_multispectral:
                t = torch.tensor(
                    np.rollaxis(orig_frame, -1, 0).astype(np.float32),
                    dtype=torch.float32,
                )[None]
                scaled_t = self._data._min_max_scaler(t)[0]
                orig_frame = (
                    (scaled_t * 255)
                    .round()
                    .numpy()
                    .astype(np.uint8)[self._data._symbology_rgb_bands]
                )
                orig_frame = np.rollaxis(orig_frame, 0, 3)
                a = np.zeros(orig_frame.shape, dtype=np.uint8)
                a[:] = orig_frame[:]
                if len(labels) > 0:
                    image = _draw_predictions(a, predictions, labels)
                else:
                    image = orig_frame
            else:
                image = _draw_predictions(orig_frame, predictions, labels)
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            plt.xticks([])
            plt.yticks([])
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return masks, predictions, labels, scores
        else:
            return masks, predictions, labels



from __future__ import annotations
import time
from typing import Any
from arcgis.geometry import Polygon, Envelope
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features import FeatureLayer

########################################################################


class ParcelFabricManager(object):
    """
    The Parcel Fabric Server is responsible for exposing parcel management
    capabilities to support a variety of workflows from different clients
    and systems.

    ====================     ====================================================================
    **Parameter**             **Description**
    --------------------     --------------------------------------------------------------------
    url                      Required String. The URI to the service endpoint.
    --------------------     --------------------------------------------------------------------
    gis                      Required :class:`~arcgis.gis.GIS`. The enterprise connection.
    --------------------     --------------------------------------------------------------------
    version                  Required :class:`~arcgis.features._version.Version`. This is the version object where the modification
                             will occur.
    --------------------     --------------------------------------------------------------------
    flc                      Required :class:`~arcgis.features.FeatureLayerCollection` . This is the parent container for
                             ParcelFabricManager.
    ====================     ====================================================================

    """

    _con = None
    _flc = None
    _gis = None
    _url = None
    _version = None
    _properties = None
    # ----------------------------------------------------------------------

    def __init__(self, url, gis, version, flc):
        """Constructor"""
        self._url = url
        self._gis = gis
        self._con = gis._portal.con
        self._version = version
        self._flc = flc

    # ----------------------------------------------------------------------

    def __str__(self):
        return "< ParcelFabricManager @ %s >" % self._url

    # ----------------------------------------------------------------------

    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------

    def __enter__(self):
        return self

    # ----------------------------------------------------------------------

    def __exit__(self, type, value, traceback):
        return

    # ----------------------------------------------------------------------

    @property
    def layer(self):
        """returns the Parcel Layer ( :class:`~arcgis.features.FeatureLayer` object or None ) for the service"""
        if (
            "controllerDatasetLayers" in self._flc.properties
            and "parcelLayerId" in self._flc.properties.controllerDatasetLayers
        ):
            url = "%s/%s" % (
                self._flc.url,
                self._flc.properties.controllerDatasetLayers.parcelLayerId,
            )
            return FeatureLayer(url=url, gis=self._gis)
        return None

    # ----------------------------------------------------------------------

    @property
    def properties(self):
        """returns the properties of the service"""
        if self._properties is None:
            res = self._con.get(self._url, {"f": "json"})
            self._properties = PropertyMap(res)
        return self._properties

    # ----------------------------------------------------------------------

    def assign_to_record(
        self,
        features: list[dict[str, Any]],
        record: str,
        write_attribute: str,
        moment: int | str | None = None,
        future: bool = False,
    ):
        """
        Assigns the specified parcel features to the specified record. If
        parcel polygons are assigned, the record polygon will be updated to
        match the cumulative geometry of all the parcels associated to it.
        The Created By Record or Retired By Record attribute field of the
        parcel features is updated with the global ID of the assigned
        record.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        features                 Required List. The parcel features to assign to the specified record.
                                 Can be parcels, parcel polygons, parcel points, and parcel lines.

                                 :Syntax:

                                 .. code-block:: python

                                     >>> features=[{"id":"<guid>","layerId":"<layerID>"},{...}]

        --------------------     --------------------------------------------------------------------
        record                   Required String. The record that will be assigned to the specified
                                 parcel features.
        --------------------     --------------------------------------------------------------------
        write_attribute          Required String. Represents the record field to update on the parcel
                                 features. Either the Created By Record or Retired By Record field is
                                 to be updated with the global ID of the assigned record.

                                 Allowed Values:

                                    `CreatedByRecord` or `RetiredByRecord`

        --------------------     --------------------------------------------------------------------
        moment                   Optional Integer. This should only be specified by the client when
                                 they do not want to use the current moment

        --------------------     --------------------------------------------------------------------
        future                   Optional boolean. If `True`, the request is processed as an asynchronous
                                 job and a URL is returned that points a location displaying the status
                                 of the job.

                                 The default is `False`.
        ====================     ====================================================================

        :return: Boolean. `True` if successful otherwise `False`

        """
        url = "{base}/assignFeaturesToRecord".format(base=self._url)
        if moment is None:
            moment = int(time.time())
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "moment": moment,
            "parcelFeatures": features,
            "record": record,
            "writeAttribute": write_attribute,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def build(
        self,
        extent: dict | Envelope | None = None,
        moment: int | str | None = None,
        return_errors: bool = False,
        record: str | None = None,
        future: bool = False,
    ):
        """
        A `build` will fix known parcel fabric errors.

        For example, if a parcel polygon exists without lines, then build will
        construct the missing lines. If lines are missing, the polygon row(s)
        are created. When constructing this objects, build will attribute the
        related keys as appropriate. Build also maintains `lineage` and `record`
        features. The parcel fabric must have sufficient information for build
        to work correctly. Ie, source reference document, and connected lines.

        Build provides options to increase performance. The process can just
        work on specific parcels, geometry types or only respond to parcel point
        movement in the case of an adjustment.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        extent                   Optional :class:`~arcgis.geometry.Envelope` . The extent to build.


                                 :Syntax:

                                 .. code-block:: python

                                     >>> extent={
                                                 "xmin":X min,
                                                 "ymin": y min,
                                                 "xmax": x max,
                                                 "ymax": y max,
                                                 "spatialReference": {"wkid": <wkid_value>}
                                                }

        --------------------     --------------------------------------------------------------------
        moment                   Optional String. This should only be specified by the client when
                                 they do not want to use the current moment
        --------------------     --------------------------------------------------------------------
        return_errors            Optional Boolean. If `True`, a verbose response will be given if errors
                                 occured.  The default is `False`.  **Deprecated**
        --------------------     --------------------------------------------------------------------
        record                   Optional String. Represents the record identifier (guid).  If a
                                 record guid is provided, only parcels associated to the record are
                                 built, regardless of the build extent.

        --------------------     --------------------------------------------------------------------
        future                   Optional boolean. If `True`, the request is processed as an asynchronous
                                 job and a URL is returned that points a location displaying the status
                                 of the job.

                                 The default is `False`.
        ====================     ====================================================================


        :return: Boolean. `True` if successful else `False`

        """
        if extent:
            extent = self._validate_extent(extent)

        url = "{base}/build".format(base=self._url)
        if moment is None:
            moment = int(time.time())
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "moment": moment,
            "buildExtent": extent,
            "record": record,
            "async": future,
            # "returnErrors" : return_errors,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def clip(
        self,
        parent_parcels: list[dict[str, Any]],
        clip_record: str | None = None,
        clipping_parcels: list[dict[str, Any]] | None = None,
        geometry: Polygon | None = None,
        moment: int | str | None = None,
        option: str | None = None,
        area_unit: str | None = None,
        future: bool = False,
    ):
        """

        Clip cuts a new child parcel into existing parent parcels. Commonly
        it retires the parent parcel(s) it cuts into to generate a reminder
        child parcel. This type of split is often part of a `parcel split
        metes and bounds` record driven workflow.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        parent_parcels              parent parcels that will be clipped into.


                                    :Syntax:

                                    .. code-block:: python

                                        >>> parent_parcels= <parcel (guid)+layer (name)...>

        -----------------------     --------------------------------------------------------------------
        clip_record                 Optional String. It is the GUID for the active legal record.
        -----------------------     --------------------------------------------------------------------
        clipping_parcels            Optional List. A list of child parcels that will be used to clip
                                    into the parent parcels. Parcel lineage is created if the child
                                    'clipping_parcels' and the parcels being clipped are of the same
                                    parcel type.


                                    :Syntax: ``clipping_parcels= <"id" : "parcel guid", "layerId": "<layer id>"...>``

                                    .. code-block:: python

                                        # Example:
                                        >>> clipping_parcels = [{"id":"{D01D3F47-5FE2-4E39-8C07-E356B46DBC78}","layerId":"16"}]

                                    .. note::
                                        Either `clipping_parcels` or `geometry` is required.

        -----------------------     --------------------------------------------------------------------
        geometry                    Optional Polygon. Allows for the clipping a parcel based on geometry instead of
                                    'clipping_parcels' geometry. No parcel lineage is created.

                                    .. note::
                                        Either `clipping_parcels` or `geometry` is required.

        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This should only be specified by the client when
                                    they do not want to use the current moment
        -----------------------     --------------------------------------------------------------------
        option                      Optional String. Represents the type of clip to perform:

                                    -  `PreserveArea` - Preserve the areas that intersect and discard the remainder areas. (default)
                                    -  `DiscardArea` - Discard the areas that intersect and preserve the remainder areas.
                                    -  `PreserveBothAreasSplit` - Preserve both the intersecting and remainder areas.
        -----------------------     --------------------------------------------------------------------
        area_unit                   Optional String. Area units to be used when calculating the stated
                                    areas of the clipped parcels. The stated area of the clipped parcels
                                    will be calculated if the stated areas exist on the parent parcels
                                    being clipped.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/clip".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parentParcels": parent_parcels,
            "moment": moment,
            "record": clip_record,
            "clippingParcels": clipping_parcels,
            "clippingGeometry": geometry,
            "clipOption": option,
            "defaultAreaUnit": area_unit,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def merge(
        self,
        parent_parcels: list[dict[str, Any]],
        target_parcel_type: str,
        attribute_overrides: dict[str, Any] | None = None,
        child_name: str | None = None,
        default_area_unit: int | None = None,
        merge_record: str | None = None,
        merge_into: str | None = None,
        moment: int | str | None = None,
        future: bool = False,
    ):
        """
        Merge combines 2 or more parent parcels into onenew child parcel. Merge
        sums up legal areas of parent parcels to the new child parcel legal
        area (using default area units as dictated by client). The child parcel
        lines arecomposed from the outer boundaries of the parent parcels.
        Merge can create multipart parcels as well as proportion lines (partial
        overlap of parent parcels). Record footprint is updated to match the
        child parcel.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        parent_parcels           Required String. It is the parcel(guid)+layer(name) identifiers to
                                 merge.
        --------------------     --------------------------------------------------------------------
        target_parcel_type       Required String. Layer where parcel is merged to.  History is
                                 created when parents and child are of the same parcel type
        --------------------     --------------------------------------------------------------------
        attribute_overrides      Optional List. A list of attributes to set on the child parcel, if
                                 they exist. Pairs of field name and value.


                                 :Syntax:

                                 .. code-block:: python

                                     >>> attribute_overrides = [
                                                               {
                                                                "type": "PropertySet",
                                                                "propertySetItems": [
                                                                                     <field name>,
                                                                                     <field value>
                                                                                    ]
                                                               }
                                                              ]

                                 .. note::
                                     To set subtype, include subtype value in this list.

        --------------------     --------------------------------------------------------------------
        child_name               Optional String. A descript of the child layer. **DEPRECATED**
        --------------------     --------------------------------------------------------------------
        default_area_unit        Optional String. The area units of the child parcel.
        --------------------     --------------------------------------------------------------------
        merge_record             Optional String. Record identifier (guid).  If missing, no history
                                 is created.
        --------------------     --------------------------------------------------------------------
        merge_into               Optional String. A parcel identifier (guid). Invalid to have a
                                 record id.
        --------------------     --------------------------------------------------------------------
        moment                   Optional String. This parameter represents the session moment (the
                                 default is the version current moment). This should only be
                                 specified by the client when they do not want to use the current
                                 moment.
        --------------------     --------------------------------------------------------------------
        area_unit                Optional Integer. Represents the default area units to be used when
                                 calculating the stated area of the merged parcel. The stated area of
                                 the merged parcel will be calculated if the stated areas exist on
                                 the parcels being merged.

        --------------------     --------------------------------------------------------------------
        future                   Optional boolean. If `True`, the request is processed as an asynchronous
                                 job and a URL is returned that points a location displaying the status
                                 of the job.

                                 The default is `False`.
        ====================     ====================================================================


        :return: Dictionary indicating 'success' or 'error'

        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/merge".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parentParcels": parent_parcels,
            "record": merge_record,
            "moment": moment,
            "targetParcelType": target_parcel_type,
            "mergeInto": merge_into,
            # "childName" : child_name,
            "defaultAreaUnit": default_area_unit,
            "attributeOverrides": attribute_overrides,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def copy_lines_to_parcel_type(
        self,
        parent_parcels: list[dict[str, Any]],
        record: str,
        target_type: int | str,
        moment: int | str | None = None,
        mark_historic: bool = False,
        use_source_attributes: bool = False,
        attribute_overrides: dict[str, Any] | None = None,
        use_polygon_attributes: bool = False,
        parcel_subtype: int | None = None,
        future: bool = False,
    ):
        """

        Copy lines to parcel type is used when the construction of the
        child parcel is based on parent parcel geometry. It creates a
        copy of the parent parcels lines that the user can modify (insert,
        delete, update) before they build the child parcels. If the source
        parcel type and the target parcel type are identical (common)
        parcel lineage is created.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        parent_parcels              Required String. Parcel parcels from which lines are copied.
        -----------------------     --------------------------------------------------------------------
        record                      Required String. The unique identifier (guid) of the active legal
                                    record.
        -----------------------     --------------------------------------------------------------------
        target_type                 Required String. The target parcel layer to which the lines will be
                                    copied to.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        mark_historic               Optional Boolean. Mark the parent parcels historic. The default is
                                    `False`.
        -----------------------     --------------------------------------------------------------------
        use_source_attributes       Optional Boolean. If the source and the target line schema match,
                                    attributes from the parent parcel lines will be copied to the new
                                    child parcel lines when it is set to  `True`. The default is `False`.
        -----------------------     --------------------------------------------------------------------
        use_polygon_attributes      Optional Boolean. Parameter representing whether to preserve and
                                    transfer attributes of the parent parcels to the generated seeds.
        -----------------------     --------------------------------------------------------------------
        attribute_overrides         Optional Dictionary. To set fields on the child parcel lines with a
                                    specific value. Uses a key/value pair of FieldName/Value.

                                    Syntax:

                                    .. code-block:: python

                                        >>> attributeOverrides = [{"type": "PropertySet",
                                                                   "propertySetItems": [
                                                                                        <field name>,
                                                                                        <field value>
                                                                                       ]
                                                                 }]

        -----------------------     --------------------------------------------------------------------
        parcel_subtype              Optional Integer. Represents the target parcel subtype.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/copyLinesToParcelType".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parentParcels": parent_parcels,
            "record": record,
            "markParentAsHistoric": mark_historic,
            "useSourceLineAttributes": use_source_attributes,
            "useSourcePolygonAttributes": use_polygon_attributes,
            "targetParcelType": target_type,
            "targetParcelSubtype": parcel_subtype,
            "attributeOverrides": attribute_overrides,
            "moment": moment,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def change_type(
        self,
        parcels: list[dict[str, Any]],
        target_type: str,
        parcel_subtype: int | str = 0,
        moment: int | str | None = None,
        future: bool = False,
    ):
        """

        Changes a set of parcels to a new parcel type. It creates new
        polygons and lines and deletes them from the source type. This
        is used when a parcel was associated in the wrong parcel type subtype
        and/or when creating multiple parcels as part of a build process.
        Example: when lot parcels are created as part of a subdivision, the
        road parcel is moved to the encumbrance (easement) parcel type.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. Parcels list that will change type
        -----------------------     --------------------------------------------------------------------
        target_type                 Required String. The target parcel layer
        -----------------------     --------------------------------------------------------------------
        target_subtype              Optional Integer. Target parcel subtype. The default is 0 meaning
                                    no subtype required.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Boolean. `True` if successful else `False`


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/changeParcelType".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parcels": parcels,
            "targetParcelType": target_type,
            "targetParcelSubtype": parcel_subtype,
            "moment": moment,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def delete(
        self,
        parcels: list[dict[str, Any]],
        moment: int | str | None = None,
        future: bool = False,
    ):
        """

        Delete a set of parcels, removing associated or unused lines, and
        connected points.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. The parcels to erase.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/deleteParcels".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parcels": parcels,
            "moment": moment,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def update_history(
        self,
        features: list[dict[str, Any]],
        record: str,
        moment: int | str | None = None,
        set_as_historic: bool = False,
        future: bool = False,
    ):
        """
        Sets the specified parcel features to current or historic using the
        specified record. If setting current parcels as historic, the
        Retired By Record field of the features is updated with the Global
        ID of the specified record. If setting historic parcels as current,
        the Created By Record field of the features is updated with the
        Global ID of the specified record.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        features                    Required List. The parcel features to be set as historic or current.
                                    Can be parcels, parcel polygons, parcel points, and parcel lines.


                                    :Syntax:

                                    .. code-block:: python

                                        >>> features=[
                                                      {"id":"<guid>",
                                                       "layerId":"<layerID>"},
                                                      {...}
                                                     ]

        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        set_as_historic             Optional Boolean.  Boolean parameter representing whether to set the
                                    features as historic (`True`). If `False`, features will be set as
                                    current.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/updateParcelHistory".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment": moment,
            "record": record,
            "setAsHistoric": set_as_historic,
            "parcelFeatures": features,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def create_seeds(
        self,
        record: str,
        moment: int | str | None = None,
        extent: dict | Envelope | None = None,
        future: bool = False,
    ):
        """

        Create seeds creates parcel seeds for closed loops of lines that
        are associated with the specified record.

        When building parcels from lines, parcel seeds are used. A parcel
        seed is the initial state or seed state of a parcel. A parcel seed
        indicates to the build process that a parcel can be built from the
        lines enclosing the seed.

        A parcel seed is a minimized polygon feature and is stored in the
        parcel type polygon feature class.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        extent                      Optional Dict/ :class:`~arcgis.geometry.Envelope` . The envelope of the extent
                                    in which to create seeds.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        if extent:
            extent = self._validate_extent(extent)

        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/createSeeds".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment": moment,
            "record": record,
            "extent": extent,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def duplicate(
        self,
        parcels: list[dict[str, Any]],
        parcel_type: int | str,
        record: str,
        parcel_subtype: int | str | None = None,
        repeat_count: int | str | None = None,
        update_field: str = None,
        start_value: int | str | None = None,
        increment_value: int | str | None = None,
        moment: int | str | None = None,
        future: bool = False,
    ):
        """
        `duplicate` allows for the cloning of parcels from a specific record.

        Parcels can be duplicated in the following ways:

        -  Duplicate to a different parcel type.
        -  Duplicate to a different subtype in the same parcel type.
        -  Duplicate to a different subtype in a different parcel type.

        Similarly, parcel seeds can be duplicated to subtypes and different parcel types.

        =======================     ====================================================================
        **Parameter**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. A list of parcels to duplicate.


                                    :Syntax:

                                    .. code-block:: python

                                        >>>> parcels=[
                                                      {"id":"<parcelguid>",
                                                       "layerId":"16"
                                                      },
                                                      {...}
                                                     ]

        -----------------------     --------------------------------------------------------------------
        parcel_type                 Required Integer. The target parcel type.
        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        parcel_subtype              Optional Integer. The target parcel subtype.  The default is 0.
        -----------------------     --------------------------------------------------------------------
        repeat_count                Optional Integer. How many times to duplicate the target parcels
        -----------------------     --------------------------------------------------------------------
        update_field                Optional String. Which incrementable field to update on the target
        -----------------------     --------------------------------------------------------------------
        start_value                 Optional Integer. What value to start on when incrementing
        -----------------------     --------------------------------------------------------------------
        increment_value             Optional Integer. How many steps to increment
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.

        -----------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        if parcel_subtype is None:
            parcel_subtype = 0
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/duplicateParcels".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment": moment,
            "record": record,
            "parcels": parcels,
            "targetParcelType": parcel_type,
            "targetParcelSubtype": parcel_subtype,
            "repeatCount": repeat_count,
            "updateField": update_field,
            "startValue": start_value,
            "incrementValue": increment_value,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def analyze_least_squares_adjustment(
        self,
        analysis_type: str = "CONSISTENCY_CHECK",
        convergence_tolerance: float = 0.05,
        parcel_features: dict[str, Any] | None = None,
        future: bool = False,
    ):
        """
        .. note::
            Least Squares Adjustment functionality introduced at version 10.8.1

        Analyzes the parcel fabric measurement network by running a least squares adjustment on the
        input parcels. A least-squares adjustment is a mathematical procedure that uses statistical
        analysis to estimate the most likely coordinates for connected points in a measurement network.

        Use :meth:`~arcgis.features._parcel.ParcelFabricManager.apply_least_squares_adjustment`
        to apply the results of a least squares adjustment to parcel fabric feature classes.

        ============================    ====================================================================
        **Parameter**                    **Description**
        ----------------------------    --------------------------------------------------------------------
        analysis_type                   Optional string. Represents the type of least squares analysis that will be run on the input parcels.

                                        * CONSISTENCY_CHECK - A free-network least-squares adjustment will be run to check dimensions on
                                          parcel lines for inconsistencies and mistakes. Fixed or weighted control points will not be
                                          used by the adjustment.

                                        * WEIGHTED_LEAST_SQUARES - A weighted least-squares adjustment will be run to compute updated
                                          coordinates for parcel points. The parcels being adjusted should connect to at least two fixed
                                          or weighted control points.

                                        The default value is CONSISTENCY_CHECK.
        ----------------------------    --------------------------------------------------------------------
        convergence_tolerance           Optional float. Represents the maximum coordinate shift expected after iterating the least squares adjustment. A least
                                        squares adjustment is run repeatedly (in iterations) until the solution converges. The solution is
                                        considered converged when maximum coordinate shift encountered becomes less than the specified convergence
                                        tolerance.

                                        The default value is 0.05 meters or 0.164 feet.
        ----------------------------    --------------------------------------------------------------------
        parcel_features                 Optional list. Represents the input parcels that will be analyzed by a least squares adjustment.


                                        :Syntax:

                                        .. code-block:: python

                                            >>> parcel_features = [{"id":"<guid>","layerId":"<layerID>"},{...}]

                                        If None, the method will analyze the entire parcel fabric.
        ----------------------------    --------------------------------------------------------------------
        future                          Optional boolean. If `True`, the request is processed as an
                                        asynchronous job and a URL is returned that points a location
                                        displaying the status of the job.

                                        The default is `False`.
        ============================    ====================================================================

        :return: Boolean. `True` if successful else `False`

        """
        url = "{base}/analyzeByLeastSquaresAdjustment".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "analysisType": analysis_type,
            "convergenceTolerance": convergence_tolerance,
            "parcelFeatures": parcel_features,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def apply_least_squares_adjustment(
        self,
        movement_tolerance: float = 0.05,
        update_attributes: bool = True,
        future: bool = False,
    ):
        """
        .. note::
            Least Squares Adjustment functionality introduced at version 10.8.1

        Applies the results of a least squares adjustment to parcel fabric feature classes. Least squares adjustment results stored
        in the AdjustmentLines and AdjustmentPoints feature classes are applied to the corresponding parcel line, connection line,
        and parcel fabric point feature classes.

        Use :meth:`~arcgis.features._parcel.ParcelFabricManager.analyze_least_squares_adjustment`
        to run a least-squares analysis on parcels and store the results in adjustment feature classes.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        movement_tolerance        Optional float. Represents the minimum allowable coordinate shift when updating parcel fabric points. If the distance between
                                  the adjustment point and the parcel fabric point is greater than the specified tolerance, the parcel fabric
                                  point is updated to the location of the adjustment point.

                                  The default tolerance is 0.05 meters or 0.164 feet.
        --------------------     --------------------------------------------------------------------
        update_attributes         Optional boolean. Specifies whether attribute fields in the parcel fabric Points feature class will be updated with
                                  statistical metadata. The XY Uncertainty, Error Ellipse Semi Major, Error Ellipse Semi Minor, and
                                  Error Ellipse Direction fields will be updated with the values stored in the same fields in the AdjustmentPoints
                                  feature class.

                                  The default is `True`
        --------------------     --------------------------------------------------------------------
        future                   Optional boolean. If `True`, the request is processed as an asynchronous job and a URL is returned that points a location
                                 displaying the status of the job.

                                 The default is `False`.
        ====================     ====================================================================

        :return: Boolean. `True` if successful else `False`

        """

        url = "{base}/applyLeastSquaresAdjustment".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "movementTolerance": movement_tolerance,
            "updateAttributes": update_attributes,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            future = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return future
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def divide(
        self,
        divide_parcel_guid: str,
        divide_parcel_type: int | str | None,
        divide_record: str,
        divide_option: str,
        divide_number_of_parts: int | str,
        divide_part_area: float | int,
        divide_line_bearing: float,
        divide_left_side: bool,
        divide_distribute_remainder: bool,
        default_area_unit: int | str | None = None,
        divide_cogo_line_bearing: float = None,
        future: bool = False,
    ):
        """
        .. note::
            Divide functionality introduced at version 10.9.1

        Divide a polygon feature into multiple features that have proportional or equal areas, or equal widths.

        =========================== ====================================================================
        **Parameter**                **Description**
        --------------------------- --------------------------------------------------------------------
        divide_parcel_guid          Required String. Parameter for the unique identifier `guid` of the
                                    parcel being divided.
        --------------------------- --------------------------------------------------------------------
        divide_parcel_type          Required Integer. Parameter representing the parcel type layer ID in
                                    which the new, divided parcels will be created.
        --------------------------- --------------------------------------------------------------------
        divide_record               Required String. Parameter for the unique identifier `guid` of the
                                    record being used for the divide.
                                    If missing, no parcel history is created.
        --------------------------- --------------------------------------------------------------------
        divide_option               Required String. The type of division to be performed.

                                    - `ProportionalArea`
                                    - `EqualArea`
                                    - `EqualWidth`
        --------------------------- --------------------------------------------------------------------
        divide_number_of_parts      Required Integer. The number parts into which the parcel will
                                    be divided.
        --------------------------- --------------------------------------------------------------------
        divide_part_area            Required Float. Area (or width) of each part (parcel fabric GDB units squared).

                                    .. note::
                                        This value is ignored when dividing by proportional area. A
                                        default value of 0 will be applied.
        --------------------------- --------------------------------------------------------------------
        divide_line_bearing         Required Float. The direction (in decimal degrees) of the line
                                    used to divide the parcel.
        --------------------------- --------------------------------------------------------------------
        divide_left_side            Required Boolean. Parameter indicating if area being divided is
                                    starting from the leftmost edge of the parcel. Any remainder area
                                    will be to the right of the divided parts. If `False`, the area being
                                    divided starts from the rightmost edge of the parcel and any remainder
                                    area will be to the left of the divided parts.

                                    This parameter is required for the `EqualArea` and `EqualWidth`
                                    divide options.

                                    .. note::
                                        This value is ignored when dividing by proportional area. A
                                        default value of `False` will be applied.
        --------------------------- --------------------------------------------------------------------
        divide_distribute_remainder Required Boolean. Indicates whether to distribute or merge the
                                    remainder area after the divide is performed. This parameter is used
                                    for the `EqualArea` and `EqualWidth` divide options.

                                    .. note::
                                        This value is ignored when dividing by proportional area. A
                                        default value of `False` will be applied.
        --------------------------- --------------------------------------------------------------------
        default_area_unit           Required Integer. The units in which area will be stored. The parameter
                                    is specified as a domain code from the `PF_AreaUnits` parcel fabric
                                    domain.

                                    .. code-block:: python

                                        #Example Usage:

                                        #Square feet
                                        >>> default_area_unit=109405
                                        #Square meters
                                        >>> default_area_unit=109404

        --------------------------- --------------------------------------------------------------------
        divide_cogo_line_bearing    Optional Float. Parameter representing the COGO direction
                                    (in decimal degrees) that will be stored in the COGO Direction field
                                    of the dividing lines.
        --------------------------- --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =========================== ====================================================================

        :return: Dictionary indicating 'success' or 'error'


        """
        if divide_option == "ProportionalArea":
            if not divide_part_area:
                divide_part_area = 0
            if not divide_left_side:
                divide_left_side = False
            if not divide_distribute_remainder:
                divide_distribute_remainder = False

        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/divide".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "divideParcelGuid": divide_parcel_guid,
            "divideParcelType": divide_parcel_type,
            "record": divide_record,
            "divideOption": divide_option,
            "divideNumberOfParts": divide_number_of_parts,
            "dividePartAreaOrWidth": divide_part_area,
            "divideLineBearing": divide_line_bearing,
            "divideLeftSide": divide_left_side,
            "divideDistributeRemainder": divide_distribute_remainder,
            "defaultAreaUnit": default_area_unit,
            "divideCogoLineBearing": divide_cogo_line_bearing,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def reassign_features_to_record(
        self,
        source_record: str,
        target_record: str,
        delete_source_record: bool,
        future: bool = False,
    ):
        """
        Reassigns all parcel features in the specified source record to the specified target record.
        The source record will become empty and will be associated to no parcel features. The record
        polygon of the target record will be updated to match the cumulative geometry of all the
        parcels associated to it.

        The Created By Record or Retired By Record attribute field of the
        parcel features is updated with the global ID of the assigned
        record.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        source_record            Required String. GlobalID representing the record containing the
                                 parcel features to be reassigned.


                                 Syntax:

                                  source_record=<guid>

        --------------------     --------------------------------------------------------------------
        target_record            Required String. GlobalID representing the target record to which
                                 the parcel features will be reassigned.


                                 Syntax

                                    target_record=<guid>

        --------------------     --------------------------------------------------------------------
        delete_source_record     Required Bool. Parameter indicating whether to delete the original
                                 source record.

        --------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        ====================     ====================================================================

        :returns: Boolean

        """
        url = "{base}/reassignFeaturesToRecord".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "sourceRecord": source_record,
            "targetRecord": target_record,
            "deleteSourceRecord": delete_source_record,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def reconstruct_from_seeds(
        self,
        extent: dict | Envelope,
        future: bool = False,
    ):
        """
        This operation constructs parcels from seeds enclosed by parcel lines in the specified extent. The tool reconstructs parcels regardless of the parcel
        lines associations with records.

        ====================     ====================================================================
        **Parameter**             **Description**
        --------------------     --------------------------------------------------------------------
        extent                   Parameter representing the envelope of the extent to reconstruct seeds.
                                 Seeds that lie within the specified extent will be reconstructed into
                                 parcels.


                                 :Syntax:

                                 .. code-block:: python

                                     >>> extent={
                                                 "xmin":X min,
                                                 "ymin": y min,
                                                 "xmax": x max,
                                                 "ymax": y max,
                                                 "spatialReference": {"wkid": <wkid_value>}
                                                }

        --------------------     --------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        ====================     ====================================================================

        :return: Dictionary indicating 'success' or 'error'

        """
        if extent:
            extent = self._validate_extent(extent)

        url = "{base}/reconstructFromSeeds".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "extent": extent,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def transfer_parcel(
        self,
        transfer_parcel_feature: dict[str, Any],
        target_parcel_features: list[dict[str, Any]],
        record: str,
        default_area_unit: int,
        source_parcel_features: list[dict[str, Any]] | None = None,
        future: bool = False,
    ):
        """
        The :meth:`~transfer_parcel` method supports workflows for transferring
        a piece of land between parcels.

        =======================     =======================================================================
        **Parameter**                **Description**
        -----------------------     -----------------------------------------------------------------------
        transfer_parcel_feature     Required Dict. Parameter representing the parcel to be transferred.
                                    Only one parcel can be specified as the transfer parcel.

                                    .. code-block:: python

                                        # Example Usage:

                                        >>> transfer_parcel_feature={"id":"<guid>","layerId":"<layerID>"}
        -----------------------     -----------------------------------------------------------------------
        target_parcel_features      Required List. Parameter representing the target parcels to which land
                                    will be transferred. These parcels will be merged with the transfer
                                    parcel and will become larger.

                                    .. code-block:: python

                                        # Example Usage:

                                        >>> target_parcel_features=[{"id":"<guid>","layerId":"<layerID>"},{...}]

        -----------------------     -----------------------------------------------------------------------
        record                      Required String. Parameter for the unique identifier (GUID) of the
                                    record being used for the transfer
        -----------------------     -----------------------------------------------------------------------
        default_area_unit           Required Integer. The units in which area will be stored. The parameter
                                    is specified as a domain code from the `PF_AreaUnits` parcel fabric
                                    domain.

                                    .. code-block:: python

                                        #Example Usage:

                                        #Square feet
                                        >>> default_area_unit=109405
                                        #Square meters
                                        >>> default_area_unit=109404

        -----------------------     -----------------------------------------------------------------------
        source_parcel_features      Optional List. Parameter representing the source parcels from which
                                    land will be transferred. These parcels will be clipped and will
                                    become smaller.

                                    .. code-block:: python

                                        # Example Usage:

                                        >>> source_parcel_features=[{"id":"<guid>","layerId":"<layerID>"},{...}]

        -----------------------     -----------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     =======================================================================

        :return: Dictionary indicating 'success' or 'error' with a list of edited features

        """
        url = "{base}/transferParcel".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "transferParcelFeature": transfer_parcel_feature,
            "sourceParcelFeatures": source_parcel_features,
            "targetParcelFeatures": target_parcel_features,
            "record": record,
            "defaultAreaUnit": default_area_unit,
            "async": future,
            "f": "json",
        }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def set_line_label_position(
        self,
        parcel_line_features: list[dict[str, Any]],
        future: bool = False,
    ):
        """
        The :meth:`~set_line_label_position` sets the label position of the line's COGO dimension to the
        left of the parcel line, to the right of the parcel line, or centered over the parcel line.

        =======================     =======================================================================
        **Parameter**                **Description**
        -----------------------     -----------------------------------------------------------------------
        parcel_line_features        Required List. Parameter representing the input parcel line layers with
                                    label positions that will be updated.

                                    .. code-block:: python

                                        >>> parcel_line_features=[{"id":"<guid>",
                                                                   "layerId":"<layerID>"},
                                                                   {...}]

        -----------------------     -----------------------------------------------------------------------
        future                      Optional boolean. If `True`, the request is processed as an asynchronous
                                    job and a URL is returned that points a location displaying the status
                                    of the job.

                                    The default is `False`.
        =======================     =======================================================================

        :return: Dictionary indicating 'success' or 'error' with a list of edited features

        """
        url = "{base}/setParcelLineLabelPosition".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "parcelFeatures": parcel_line_features,
            "async": future,
            "f": "json",
        }

        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(
                self._status_via_url,
                con=self._con,
                url=res["statusUrl"],
                params={"f": "json"},
            )
            return f
        else:
            return self._con.post(url, params)

    # ----------------------------------------------------------------------

    def _validate_extent(self, extent: dict | Envelope):
        """Check for valid Extent object or None"""
        from arcgis.geometry import Envelope

        if isinstance(extent, (dict, Envelope)):
            return dict(extent)
        elif extent is None:
            return None
        elif not extent is None:
            raise ValueError("Parameter `extent` must be None, Envelope or dict.")

    # ----------------------------------------------------------------------

    def _run_async(self, fn, **inputs):
        """runs the inputs asynchronously"""
        import concurrent.futures

        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=fn, **inputs)
        except:
            future = tp.submit(fn, **inputs)
        tp.shutdown(False)
        return future

    # ----------------------------------------------------------------------

    def _status_via_url(self, con, url, params):
        """
        performs the asynchronous check to see if the operation finishes
        """
        status_allowed = [
            "esriJobSubmitted",
            "esriJobWaiting",
            "esriJobExecuting",
            "esriJobSucceeded",
            "esriJobFailed",
            "esriJobTimedOut",
            "esriJobCancelling",
            "esriJobCancelled",
        ]
        status = con.get(url, params)
        while (
            status["status"] in status_allowed
            and status["status"] != "esriJobSucceeded"
        ):
            if status["status"] == "esriJobSucceeded":
                return status
            elif status["status"] in [
                "esriJobFailed",
                "esriJobTimedOut",
                "esriJobCancelled",
            ]:
                break
            status = con.get(url, params)
        return status



"""
This file can contain various metrics for different types of deep learning models. 
"""

# MultiLabel Image Classification


def flatten_check(inp, targ):
    "Check that `out` and `targ` have the same number of elements and flatten them."
    inp, targ = inp.contiguous().view(-1), targ.contiguous().view(-1)
    assert len(inp) == len(targ)
    return inp, targ


def accuracy_multi(inp, targ, thresh=0.5, sigmoid=True):
    "Compute accuracy when `inp` and `targ` are the same size."
    inp, targ = flatten_check(inp, targ)
    if sigmoid:
        inp = inp.sigmoid()
    return ((inp > thresh) == targ.bool()).float().mean()



from .api import EsriSession
from ._auth import (
    EsriAPIKeyAuth,
    EsriBasicAuth,
    EsriBuiltInAuth,
    EsriGenTokenAuth,
    EsriKerberosAuth,
    EsriNotebookAuth,
    EsriOAuth2Auth,
    EsriPKIAuth,
    EsriUserTokenAuth,
    EsriWindowsAuth,
    ArcGISProAuth,
    BaseEsriAuth,
    EsriPKCEAuth,
    EsriHttpNtlmAuth,
)
from ._error import EsriHttpResponseError
from ._version import __version__

__all__ = [
    "EsriSession",
    "EsriAPIKeyAuth",
    "EsriBasicAuth",
    "EsriBuiltInAuth",
    "EsriGenTokenAuth",
    "EsriKerberosAuth",
    "EsriNotebookAuth",
    "EsriOAuth2Auth",
    "EsriPKIAuth",
    "EsriUserTokenAuth",
    "EsriWindowsAuth",
    "ArcGISProAuth",
    "BaseEsriAuth",
    "EsriPKCEAuth",
    "__version__",
    "EsriHttpNtlmAuth",
    "EsriHttpResponseError",
]



from ._overlay import overlay
from ._select import select
from ._ops import contains, crosses
from ._ops import disjoint, equals
from ._ops import overlaps, touches
from ._ops import within

__all__ = [
    "overlay",
    "select",
    "contains",
    "crosses",
    "disjoint",
    "equals",
    "overlaps",
    "touches",
    "within",
]



from __future__ import annotations
from arcgis.gis import GIS
from .containers import ContainerManager
from .instpref import InstancePreference
from .runtime import RuntimeManager
from .snapshot import SnapshotManager
from .notebook import NotebookManager
from typing import TypeVar

__all__ = ["AGOLNotebookManager"]

K = TypeVar("K")
V = TypeVar("V")


class AGOLNotebookManager:
    _container = None
    _istpref = None
    _runtimes = None
    _snapshot = None
    _nbm = None

    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    @property
    def containers(self) -> ContainerManager:
        """
        Provides the ability to manage containers and the notebooks within them
        :returns: ContainerManager
        """
        if self._container is None:
            self._container = ContainerManager(
                url=f"{self._url}/system/containers", gis=self._gis
            )
        return self._container

    @property
    def instance_preferences(self) -> InstancePreference:
        """
        Provides information about the available instances for notebooks

        :returns: InstancePreference
        """
        if self._istpref is None:
            url = f"{self._url}/notebooks/instancePreferences"
            self._istpref = InstancePreference(url=url, gis=self._gis)
        return self._istpref

    @property
    def _machines(self) -> dict[K, V]:
        """Returns information about the machines running on notebook server"""
        url = f"{self._url}/machines"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    @property
    def runtimes(self) -> RuntimeManager:
        """Provides information about the available runtimes on the notebook server"""
        if self._runtimes is None:
            url = f"{self._url}/notebooks/runtimes"
            self._runtimes = RuntimeManager(url=url, gis=self._gis)
        return self._runtimes

    @property
    def snaphots(self) -> SnapshotManager:
        """
        Returns tools to work with snapshots on notebooks

        :returns: SnapshotManager
        """
        if self._snapshot is None:
            url = f"{self._url}/notebooks/snapshots"
            self._snapshot = SnapshotManager(url=url, gis=self._gis)
        return self._snapshot

    @property
    def notebooksmanager(self) -> NotebookManager:
        """
        Manages the run and execution of notebooks

        :returns: NotebookManager
        """
        if self._nbm is None:
            url = f"{self._url}/notebooks"
            self._nbm = NotebookManager(url=url, gis=self._gis, nbs=self)
        return self._nbm
